#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2022 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

"""Parse IF_DATA sections with dynamically created AML grammar.

Note
----
Currently there is no resonable error-handling / -recovery.
"""

from collections import defaultdict
from logging import getLogger

import antlr4
from antlr4.BufferedTokenStream import BufferedTokenStream
from antlr4.error.ErrorListener import ErrorListener

# import importlib as imp
# import pya2l
# imp.reload(pya2l)
# import pya2l.a2llg
# imp.reload(pya2l.a2llg)
# import pya2l.aml
# imp.reload(pya2l.aml)
# import pya2l.aml.classes
# imp.reload(pya2l.aml.classes)

from pya2l import a2llg
from pya2l.aml.classes import AMLPredefinedTypes, Referrer


EOF = -1


class EOFReached(Exception):
    """Signals end of token stream."""


IF_DATA_ELEMENTS = (
    "AXIS_PTS",
    "CHARACTERISTIC",
    "FRAME",
    "FUNCTION",
    "GROUP",
    "MEASUREMENT",
    "MEMORY_LAYOUT",
    "MEMORY_SEGMENT",
    "MODULE",
)


def token_to_int(token):
    """ """
    text = token.text
    if text.startswith(("0x", "0X")):
        return int(text, 16)
    else:
        return int(text, 10)


def token_to_float(token):
    """ """
    return float(token.text)


class IfDataParser:
    """ """

    def __init__(self, syntax):
        self.logger = getLogger("pya2l.if_data_parser")
        self.syntax = syntax

    @property
    def current_element(self):
        return self.elements[-1]

    def push_element(self, element):
        self.elements.append(element)

    def pop_element(self):
        return self.elements.pop(-1)

    def empty_stack(self):
        self.elements = []

    def parse(self, buffer):
        """ """
        self.empty_stack()
        self.push_element(self.syntax.root_element)
        input_stream = antlr4.InputStream(buffer)
        self.lexer = a2llg.a2llg(input_stream)
        self.tokens = [t for t in self.lexer.getAllTokens() if t.channel != t.HIDDEN_CHANNEL]
        self.num_tokens = len(self.tokens)
        self.token_idx = 0
        self.level = 0
        if self.current_element.is_block:
            result = self.block()
        else:
            raise ValueError("Expected block statement.")
        return result

    def enter(self, name):
        self.level += 1

    def leave(self, name):
        self.level -= 1

    @property
    def current_token(self):
        """Get the token at the current stream position."""
        return self.lookahead(0)

    def lookahead(self, n=1):
        """Get the token `n` elements ahead of current stream position."""
        index = self.token_idx + n
        if index < self.num_tokens:
            return self.tokens[index]
        else:
            raise EOFReached()

    def value(self, token):
        """Convert token text to Python type."""
        value = None
        if token.type == self.lexer.INT:
            value = token_to_int(token)
        elif token.type == self.lexer.IDENT:
            value = token.text
        elif token.type == self.lexer.FLOAT:
            value = token_to_float(token)
        elif token.type == self.lexer.STRING:
            value = token.text[1:-1]
        return value

    def consume(self):
        """Increment token stream position by one."""
        self.token_idx += 1

    def match(self, token_type, value=None):
        ok = self.current_token.type == token_type
        token_value = self.value(self.current_token)
        self.consume()
        if value is None:
            return ok
        else:
            if not ok:
                return False
            return token_value == value

    def block(self):
        result = None
        self.enter("block")
        elem = self.current_element
        tag = elem.tag
        self.match(self.lexer.BEGIN)
        self.match(self.lexer.IDENT, elem.tag)
        tp = elem.type_name.type_
        if isinstance(tp, Referrer):
            entry = self.syntax.get_type(tp.category, tp.identifier)
        else:
            entry = elem.definition
        self.push_element(entry)
        class_name = entry.__class__.__name__
        if class_name == "TaggedUnion":
            result = {"tag": tag, "value": self.tagged_union()}
        elif class_name == "StructType":
            result = {"tag": tag, "value": self.struct()}
        elif class_name == "TaggedStructType":
            result = {"tag": tag, "value": self.tagged_struct()}
        elif class_name == "PredefinedType":
            result = {"tag": tag, "value": self.predefined_type()}
        elif class_name == "Enumeration":
            result = {"tag": tag, "value": self.enum()}
        else:
            raise NotImplementedError(entry.__class__.__name__)
        self.pop_element()
        self.match(self.lexer.END)
        self.match(self.lexer.IDENT, elem.tag)
        self.leave("block")
        return result

    def tagged_union(self):
        result = []
        self.enter("tagged_union")
        tag = self.current_token.text
        # name = self.current_element.name
        definition = self.current_element.tags.get(tag)
        if definition:
            tp = definition.type_name
            tn = tp.type_.__class__.__name__
            self.consume()
            self.push_element(tp.type_)
            if tn == "StructType":
                result = self.struct()
            elif tn == "TaggedStructType":
                result = self.tagged_struct()
            elif tn == "TaggedUnion":
                result = self.tagged_union()
            elif tn == "PredefinedType":
                result = self.predefined_type()
            elif tn == "Enumeration":
                result = self.enum()
            else:
                raise NotImplementedError(tn)
            self.pop_element()
        self.leave("tagged_union")
        return result

    def predefined_type(self):
        self.enter("predefined_type")
        arr = self.current_element.array_specifier
        tp = self.current_element.type_
        if arr:
            if tp in (AMLPredefinedTypes.PDT_CHAR, AMLPredefinedTypes.PDT_UCHAR):
                result = self.value(self.current_token)
                self.consume()
            else:
                result = []
                for _ in range(arr[0]):
                    result.append(self.value(self.current_token))
                    self.consume()
        else:
            result = self.value(self.current_token)
            self.consume()
        self.leave("predefined_type")
        return result

    def enum(self):
        self.enter("enum")
        enumerator = self.current_token.text
        self.match(self.lexer.IDENT)
        # ok = enumerator in self.current_element.enumerators
        self.leave("enum")
        return enumerator

    def struct(self):
        self.enter("struct")
        members = self.current_element.members
        # name = self.current_element.name
        # print("struct: ", name)
        result = []
        for mem in members:
            value = mem.value
            # multiple = mem.multiple
            entry = value.type_name.type_
            # idi = entry.identifier if hasattr(entry, "identifier") else "n/a"
            if isinstance(entry, Referrer):
                entry = self.syntax.get_type(entry.category, entry.identifier)
            self.push_element(entry)
            tn = entry.__class__.__name__
            if tn == "TaggedStructType":
                result.append(self.tagged_struct())
            elif tn == "StructType":
                result.append(self.struct())
            elif tn == "TaggedUnion":
                result.append(self.tagged_union())
            elif tn == "PredefinedType":
                result.append(self.predefined_type())
            elif tn == "Enumeration":
                result.append(self.enum())
            else:
                raise NotImplementedError(tn)
            self.pop_element()
        self.leave("struct")
        return result

    def block_or_tag(self):
        if self.current_token.type == self.lexer.BEGIN:
            return self.lookahead(1).text
        else:
            return self.current_token.text

    def tagged_struct(self):
        self.enter("tagged_struct")
        tag = self.block_or_tag()
        tags = self.current_element.tags.keys()
        counter = {k: False for k in tags}
        tag_dict = defaultdict(list)
        while True:
            definition = self.current_element.tags.get(tag)
            if definition:
                counter[tag] = True
                if definition.block_definition:
                    self.push_element(definition.block_definition)
                    block = self.block()
                    tag = block.pop("tag")
                    value = block.get("value")
                    if definition.multiple:
                        tag_dict[tag].append(value)
                    else:
                        tag_dict[tag] = value
                    self.pop_element()
                    if ((self.current_token.type != self.lexer.IDENT) or (self.current_token.text not in tags)) and (
                        self.current_token.type != self.lexer.BEGIN
                    ):
                        break
                elif definition.taggedstruct_definition.member:
                    self.consume()
                    if definition.taggedstruct_definition.__class__.__name__ == "TaggedStructDefinition":
                        self.push_element(definition.taggedstruct_definition)
                        tsm = self.tagged_struct_member()
                        tag = tsm.pop("tag")
                        value = tsm.get("value")
                        if definition.multiple:
                            tag_dict[tag].append(value)
                        else:
                            tag_dict[tag] = value
                        self.pop_element()
                        if ((self.current_token.type != self.lexer.IDENT) or (self.current_token.text not in tags)) and (
                            self.current_token.type != self.lexer.BEGIN
                        ):
                            break
                    else:
                        raise NotImplementedError(definition.taggedstruct_definition.__class__.__name__)
                else:
                    tag_dict[tag] = None  # just a tag.
                    self.consume()
                tag = self.block_or_tag()
            else:
                break
        self.leave("tagged_struct")
        return dict(tag_dict)

    def tagged_struct_member(self):
        tp = self.current_element.member.type_name.type_
        tag = self.current_element.tag
        self.push_element(tp)
        class_name = tp.__class__.__name__
        if class_name == "PredefinedType":
            result = {"value": self.predefined_type()}
        elif class_name == "StructType":
            result = {"type": "struct", "value": self.struct()}
        elif class_name == "Enumeration":
            result = {"value": self.enum()}
        elif class_name == "TaggedUnion":
            result = {"type": "tagged_union", "value": self.tagged_union()}
        elif class_name == "TaggedStructType":
            ts = self.tagged_struct()
            if ts:
                result = {"type": "tagged_struct", "value": ts}
        else:
            raise NotImplementedError(tp.__class__.__name__)
        self.pop_element()
        return_value = {"tag": tag}
        return_value.update(result)
        return return_value

    def type_as_str(self, type_):
        TYPE_MAP = {
            self.lexer.IDENT: "IDENT",
            self.lexer.FLOAT: "FLOAT",
            self.lexer.INT: "INT",
            self.lexer.COMMENT: "COMMENT",
            self.lexer.WS: "WS",
            self.lexer.STRING: "STRING",
            self.lexer.BEGIN: "BEGIN",
            self.lexer.END: "END",
        }
        return TYPE_MAP.get(type_)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Classes and factories regarding AML abstract syntax tree."""

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2021 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.1.0"


import enum
import json

##
## Model Classes.
##


class Element:
    def __init__(self):
        pass


class Referrer:
    def __init__(self, category, identifier):
        self.category = category
        self.identifier = identifier

    def __str__(self):
        return "Referrer(category = '{}', identifier = '{}')".format(self.category, self.identifier)

    __repr__ = __str__


class Sequence:
    def __init__(self, elements: list):
        self._current_pos = 0
        self.elements = elements

    @property
    def current_pos(self):
        return self._current_pos


class AMLPredefinedTypes(enum.IntEnum):
    """"""

    PDT_CHAR = 0
    PDT_INT = 1
    PDT_LONG = 2
    PDT_UCHAR = 3
    PDT_UINT = 4
    PDT_ULONG = 5
    PDT_DOUBLE = 6
    PDT_FLOAT = 7


def map_predefined_type(name):
    MAP = {
        "char": AMLPredefinedTypes.PDT_CHAR,
        "int": AMLPredefinedTypes.PDT_INT,
        "long": AMLPredefinedTypes.PDT_LONG,
        "uchar": AMLPredefinedTypes.PDT_UCHAR,
        "uint": AMLPredefinedTypes.PDT_UINT,
        "ulong": AMLPredefinedTypes.PDT_ULONG,
        "double": AMLPredefinedTypes.PDT_DOUBLE,
        "float": AMLPredefinedTypes.PDT_FLOAT,
    }
    return MAP.get(name)


class BaseType:
    """ """

    _is_block = False

    @staticmethod
    def _try(o):
        try:
            return o.__dict__
        except Exception:
            return str(o)

    @property
    def is_block(self):
        return self._is_block

    def to_json(self):
        return json.dumps(
            self,
            default=lambda o: self._try(o),
            sort_keys=True,
            indent=4,
            separators=(",", ":"),
        ).replace("\n", "")


class Enumeration(BaseType):
    def __init__(self, name, enumerators):
        self.name = name
        self.enumerators = {e.tag: e.constant for e in enumerators}
        self._renumber_constants()

    def __repr__(self):
        return "Enumeration(name = '{}', enumerators = {})".format(self.name, self.enumerators)

    def __contains__(self, name):
        return name in self.enumerators

    def _renumber_constants(self):
        """ISO C/C++ like enumerator numbering.

        Note
        ----
        Descending orderered ``enum``s are not supported yet.
        """
        last_idx = 0
        for tag, value in self.enumerators.items():
            if value is None:
                self.enumerators[tag] = last_idx
                last_idx += 1
            else:
                last_idx = value + 1

class Enumerator(BaseType):
    def __init__(self, tag, constant):
        self.tag = tag
        self.constant = constant

    def __repr__(self):
        return "Enumerator(tag = {}, constant = {})".format(self.tag, self.constant)


class TaggedUnion(BaseType):
    def __init__(self, name, members):
        self.name = name
        self.members = members
        self.tags = {}
        for mem in members:
            if mem.member:
                key = mem.tag
                value = mem.member
            elif mem.block_definition:
                key = mem.block_definition.tag
                value = mem.block_definition
            self.tags[key] = value

    def __repr__(self):
        return "TaggedUnion(name = '{}', tags = {}, members = {})".format(self.name, sorted(list(self.tags)), self.members)


class TaggedUnionMember(BaseType):
    def __init__(self, tag, member, block_definition):
        self.tag = tag
        self.member = member
        self.block_definition = block_definition

    def __repr__(self):
        return "TaggedUnionMember(tag = {}, member = {}, block_definition = {})".format(
            self.tag, self.member, self.block_definition
        )


class Member(BaseType):
    def __init__(self, type_name, array_specifier):
        self.type_name = type_name
        self.array_specifier = array_specifier
        if type_name.type_.__class__.__name__ == "PredefinedType":
            self.type_name.type_.array_specifier = array_specifier

    def __repr__(self):
        return "Member(type_name = {}{})".format(self.type_name, self.array_specifier or "")


class TypeName(BaseType):
    def __init__(self, tag, name, type_):
        self.tag = tag
        self.name = name
        self.type_ = type_

    def __repr__(self):
        return "TypeName(tag = {}, name = {}, type = {})".format(self.tag, self.name, self.type_)


class PredefinedType(BaseType):
    def __init__(self, type_):
        self.type_ = type_

    def __repr__(self):
        return "PredefinedType(type = {})".format(self.type_.name)


class StructType(BaseType):
    def __init__(self, name, members):
        self.name = name
        self.members = members

    def __repr__(self):
        return "StructType(name = '{}', members = {})".format(self.name, self.members)


class TaggedStructType(BaseType):
    def __init__(self, name, members):
        self.name = name
        self.members = members
        self.tags = {}
        for mem in members:
            if mem.taggedstruct_definition:
                self.tags[mem.taggedstruct_definition.tag] = mem
            elif mem.block_definition:
                self.tags[mem.block_definition.tag] = mem

    def __repr__(self):
        return "TaggedStructType(name = '{}', tags = {}, members = {})".format(
            self.name, list(sorted(self.tags.keys())), self.members
        )


class TaggedStructDefinition(BaseType):
    def __init__(self, tag, member, multiple):
        self.tag = tag
        self.member = member
        self.multiple = multiple

    def __repr__(self):
        return "TaggedStructDefinition(tag = {}, member = {}, multiple = {})".format(self.tag, self.member, self.multiple)


class StructMember(BaseType):
    def __init__(self, value, multiple):
        self.value = value
        self.multiple = multiple

    def __repr__(self):
        return "StructMember(value = {}, multiple = {})".format(self.value, self.multiple)


class TaggedStructMember(BaseType):
    def __init__(self, taggedstruct_definition, block_definition, multiple):
        self.taggedstruct_definition = taggedstruct_definition
        self.block_definition = block_definition
        self.multiple = multiple

    def __repr__(self):
        return "TaggedStructMember(taggedstruct_definition = {}, block_definition = {}, multiple = {})".format(
            self.taggedstruct_definition, self.block_definition, self.multiple
        )


class Declaration(BaseType):
    def __init__(self, block_definition, type_definition):
        self.block_definition = block_definition
        self.type_definition = type_definition

    def __repr__(self):
        return "Declaration(block_definition = {}, type_definition = {})".format(self.block_definition, self.type_definition)


class BlockDefinition(BaseType):

    _is_block = True

    def __init__(self, tag, type_name, member, multiple):
        self.tag = tag
        self.type_name = type_name
        self.member = member
        self.multiple = multiple

    @property
    def type_(self):
        return self.type_name.type_.__class__.__name__

    @property
    def definition(self):
        return self.type_name.type_

    def __repr__(self):
        return "BlockDefinition(tag = {}, type_name = {}, member = {} multiple = {})".format(
            self.tag, self.type_name, self.member, self.multiple
        )


class TypeDefinition(BaseType):
    def __init__(self, type_name):
        self.type_name = type_name

    def __repr__(self):
        return "TypeDefinition(type_name = {})".format(self.type_name)


##
## Factories.
##
def createEnumeration(name, enumerators, is_referrer=False):
    if is_referrer:
        return Referrer("Enumeration", name)
    else:
        return Enumeration(name, enumerators)


def createEnumerator(tag, constant):
    return Enumerator(tag, constant)


def createTaggedUnion(name, members, is_referrer=False):
    if is_referrer:
        return Referrer("TaggedUnion", name)
    else:
        return TaggedUnion(name, members)


def createTaggedUnionMember(tag, member, block_definition):
    return TaggedUnionMember(tag, member, block_definition)


def createMember(type_name, array_specifier):
    return Member(type_name, array_specifier)


def createTypeName(tag, name, type_):
    return TypeName(tag, name, type_)


def createPredefinedType(name):
    return PredefinedType(map_predefined_type(name))


def createStructType(name, members, is_referrer=False):
    if is_referrer:
        return Referrer("StructType", name)
    else:
        return StructType(name, members)


def createTaggedStructType(name, members, is_referrer=False):
    if is_referrer:
        return Referrer("TaggedStructType", name)
    else:
        return TaggedStructType(name, members)


def createTaggedStructDefinition(tag, member, multiple):
    return TaggedStructDefinition(tag, member, multiple)


def createStructMember(value, multiple):
    return StructMember(value, multiple)


def createTaggedStructMember(taggedstruct_definition, block_definition, multiple):
    return TaggedStructMember(taggedstruct_definition, block_definition, multiple)


def createDeclaration(block_definition, type_definition):
    return Declaration(block_definition, type_definition)


def createBlockDefinition(tag, type_name, member, multiple):
    return BlockDefinition(tag, type_name, member, multiple)


def createTypeDefinition(type_name):
    return TypeDefinition(type_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2016 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.1.0"


import unittest

from pya2l import aml

"""
"""

AML = """
/begin A2ML
  enum mem_typ   { "INTERN" = 0, "EXTERN" = 1 };
  enum addr_typ  { "BYTE" = 1, "WORD" = 2, "LONG" = 4 };
  enum addr_mode  { "DIRECT" = 0, "INDIRECT" = 1 };

  taggedunion IF_DATA {
    "DIM" taggedstruct {
    (block "SOURCE" struct  {
        struct {
            char [101];
            int;
            long;
        };
        taggedstruct {
            block "QP_BLOB" struct  {
              ulong;
              int;
              ulong;
              long;
            };
        };
    })*;

    block "TP_BLOB" struct   {
        int;
    };

    block "KP_BLOB" struct {
        ulong;
        enum addr_typ;
    };

    block "DP_BLOB" struct {
        enum mem_typ;
    };

    block "PA_BLOB" struct {
        enum addr_mode;
    };
};
};
/end A2ML
"""


class TestAcceptance(unittest.TestCase):
    def setUp(self):
        self.parser = aml.ParserWrapper("aml", "amlFile")

    def testBasicAcceptance(self):
        self.parser.parseFromString(AML)
        self.assertEqual(self.parser.numberOfSyntaxErrors, 0)


def main():
    unittest.main()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2022 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.1.0"

from decimal import Decimal as D

import antlr4

from pya2l.aml import classes


class ResultType:
    """ """

    def __init__(self, declarations, types, root_element):
        self._declarations = declarations
        self._types = types
        self._root_element = root_element

    def get_type(self, tp, name):
        tp_ = self._types.get(tp)
        if tp_:
            return tp_.get(name).type_
        else:
            return None

    @property
    def root_element(self):
        return self._root_element

    @property
    def declarations(self):
        return self._declarations

    @property
    def types(self):
        return self._types


##
## Listener.
##
class AMLListener(antlr4.ParseTreeListener):
    def __init__(self, prepro_result):
        super().__init__()
        self.root_element = None
        self.declarations = []
        self._types = dict(StructType={}, TaggedUnion={}, TaggedStructType={}, Enumeration={})

    def exitType_name(self, ctx):
        if ctx.pr:
            tp = ctx.pr.value  # Predefined type.
        elif ctx.st:
            tp = ctx.st.value  # Struct.
        elif ctx.ts:
            tp = ctx.ts.value  # TaggedStruct.
        elif ctx.tu:
            tp = ctx.tu.value  # TaggedUnion.
        elif ctx.en:
            tp = ctx.en.value  # Enum.
        else:
            pass
        try:
            name = tp.name
        except Exception:
            name = None
        tag = ctx.t.value if ctx.t else None
        ctx.value = classes.createTypeName(tag, name, tp)

    def exitPredefined_type_name(self, ctx):
        ctx.value = classes.createPredefinedType(ctx.name.text)

    def exitStruct_type_name(self, ctx):
        is_referrer = False
        if ctx.t0:
            name = ctx.t0.value
        elif ctx.t1:
            is_referrer = True
            name = ctx.t1.value
        else:
            name = None
        members = [member.value for member in ctx.l]
        value = classes.createStructType(name, members, is_referrer)
        ctx.value = value

    def exitStruct_member(self, ctx):
        if self.get_text(ctx.m0) == "*":
            value = ctx.mstar.value
            multiple = True
        else:
            value = ctx.m.value
            multiple = False
        ctx.value = classes.createStructMember(value, multiple)

    def exitTaggedstruct_type_name(self, ctx):
        is_referrer = False
        if ctx.t0:
            name = ctx.t0.value
        elif ctx.t1:
            is_referrer = True
            name = ctx.t1.value
        else:
            name = None
        members = [member.value for member in ctx.l]
        ctx.value = classes.createTaggedStructType(name, members, is_referrer)

    def exitTaggedstruct_member(self, ctx):
        if ctx.ts0:
            taggedstructDefinition = ctx.ts0.value
        elif ctx.ts1:
            taggedstructDefinition = ctx.ts1.value
        else:
            taggedstructDefinition = None
        if ctx.bl0:
            blockDefinition = ctx.bl0.value
        elif ctx.bl1:
            blockDefinition = ctx.bl1.value
        else:
            blockDefinition = None
        chs = list(ctx.getChildren())
        mult = len(chs) == 5
        ctx.value = classes.createTaggedStructMember(taggedstructDefinition, blockDefinition, mult)

    def exitTaggedstruct_definition(self, ctx):
        chs = list(ctx.getChildren())
        mult = len(chs) > 4
        tag = ctx.tag.value if ctx.tag else None
        member = ctx.mem.value if ctx.mem else None
        ctx.value = classes.createTaggedStructDefinition(tag, member, mult)

    def exitEnumerator(self, ctx):
        tag = ctx.t.value
        constant = ctx.c.value if ctx.c else None
        ctx.value = classes.createEnumerator(tag, constant)

    def exitArray_specifier(self, ctx):
        ctx.value = ctx.c.value

    def exitEnumerator_list(self, ctx):
        ctx.value = [x.value for x in ctx.ids]

    def exitEnum_type_name(self, ctx):
        is_referrer = False
        elements = ctx.l.value if ctx.l else []
        if ctx.t0:
            name = ctx.t0.value
        elif ctx.t1:
            name = ctx.t1.value
            is_referrer = True
        else:
            name = None
        ctx.value = classes.createEnumeration(name, elements, is_referrer)

    def exitType_definition(self, ctx):
        ctx.value = classes.createTypeDefinition(ctx.type_name().value)
        tn = ctx.value.type_name
        self._types[tn.type_.__class__.__name__][tn.name] = tn

    def exitMember(self, ctx):
        typename = ctx.t.value
        arraySpecifier = [a.value for a in ctx.a]
        ctx.value = classes.createMember(typename, arraySpecifier)

    def exitTagged_union_member(self, ctx):
        tag = ctx.t.value if ctx.t else None
        member = ctx.m.value if ctx.m else None
        blockDefinition = ctx.b.value if ctx.b else None
        ctx.value = classes.createTaggedUnionMember(tag, member, blockDefinition)

    def exitTaggedunion_type_name(self, ctx):
        is_referrer = False
        if ctx.t0:
            name = ctx.t0.value
        elif ctx.t1:
            name = ctx.t1.value
            is_referrer = True
        else:
            name = None
        members = [member.value for member in ctx.l]
        ctx.value = classes.createTaggedUnion(name, members, is_referrer)

    def exitBlock_definition(self, ctx):
        tag = ctx.tag.value
        typename = ctx.tn.value if ctx.tn else None
        member = ctx.mem.value if ctx.mem else None
        mult = True if self.get_text(ctx.mult) == "*" else False
        # chs = list(ctx.getChildren())
        ctx.value = classes.createBlockDefinition(tag, typename, member, mult)
        if tag == "IF_DATA":
            self.root_element = ctx.value

    def exitDeclaration(self, ctx):
        blockDefinition = ctx.b.value if ctx.b else None
        typeDefinition = ctx.t.value if ctx.t else None
        ctx.value = classes.createDeclaration(blockDefinition, typeDefinition)

    def exitAmlFile(self, ctx):
        declarations = [d.value for d in ctx.d]
        ctx.value = declarations
        self.value = declarations
        self.declarations = declarations

    def exitIntValue(self, ctx):
        ctx.value = int(ctx.i.text) if ctx.i else None

    def exitFloatValue(self, ctx):
        ctx.value = float(ctx.f.text) if ctx.f else None

    def exitNumber(self, ctx):
        ctx.value = ctx.i.value if ctx.i else ctx.f.value

    def exitStringValue(self, ctx):
        ctx.value = ctx.s.text.strip('"') if ctx.s else None

    def exitTagValue(self, ctx):
        ctx.value = ctx.s.text.replace('"', "") if ctx.s else None

    def exitIdentifierValue(self, ctx):
        ctx.value = ctx.i.text if ctx.i else None

    def exitNumericValue(self, ctx):
        if ctx.i:
            value = int(ctx.i.text)
        elif ctx.h:
            value = int(ctx.h.text, 16)
        elif ctx.f:
            value = D(ctx.f.text)
        ctx.value = value

    def get_text(self, obj):
        return obj.text if obj is not None else None

    def result(self):
        return ResultType(self.declarations, self._types, self.root_element)



#!/bin/env python

literalNames = [
    "<INVALID>",
    "'ALIGNMENT_BYTE'",
    "'ALIGNMENT_FLOAT16_IEEE'",
    "'ALIGNMENT_FLOAT32_IEEE'",
    "'ALIGNMENT_FLOAT64_IEEE'",
    "'ALIGNMENT_INT64'",
    "'ALIGNMENT_LONG'",
    "'ALIGNMENT_WORD'",
    "'ANNOTATION'",
    "'ANNOTATION_LABEL'",
    "'ANNOTATION_ORIGIN'",
    "'ANNOTATION_TEXT'",
    "'BIT_MASK'",
    "'BYTE_ORDER'",
    "'CALIBRATION_ACCESS'",
    "'CALIBRATION'",
    "'NO_CALIBRATION'",
    "'NOT_IN_MCD_SYSTEM'",
    "'OFFLINE_CALIBRATION'",
    "'DEFAULT_VALUE'",
    "'DEPOSIT'",
    "'ABSOLUTE'",
    "'DIFFERENCE'",
    "'DISCRETE'",
    "'DISPLAY_IDENTIFIER'",
    "'ECU_ADDRESS_EXTENSION'",
    "'EXTENDED_LIMITS'",
    "'FORMAT'",
    "'FUNCTION_LIST'",
    "'GUARD_RAILS'",
    "'IF_DATA'",
    "'MATRIX_DIM'",
    "'MAX_REFRESH'",
    "'MONOTONY'",
    "'MON_DECREASE'",
    "'MON_INCREASE'",
    "'STRICT_DECREASE'",
    "'STRICT_INCREASE'",
    "'MONOTONOUS'",
    "'STRICT_MON'",
    "'NOT_MON'",
    "'PHYS_UNIT'",
    "'READ_ONLY'",
    "'REF_CHARACTERISTIC'",
    "'REF_MEMORY_SEGMENT'",
    "'REF_UNIT'",
    "'STEP_SIZE'",
    "'SYMBOL_LINK'",
    "'VERSION'",
    "'ASAP2_VERSION'",
    "'A2ML_VERSION'",
    "'PROJECT'",
    "'HEADER'",
    "'PROJECT_NO'",
    "'MODULE'",
    "'A2ML'",
    "'AXIS_PTS'",
    "'CHARACTERISTIC'",
    "'ASCII'",
    "'CURVE'",
    "'MAP'",
    "'CUBOID'",
    "'CUBE_4'",
    "'CUBE_5'",
    "'VAL_BLK'",
    "'VALUE'",
    "'AXIS_DESCR'",
    "'CURVE_AXIS'",
    "'COM_AXIS'",
    "'FIX_AXIS'",
    "'RES_AXIS'",
    "'STD_AXIS'",
    "'AXIS_PTS_REF'",
    "'CURVE_AXIS_REF'",
    "'FIX_AXIS_PAR'",
    "'FIX_AXIS_PAR_DIST'",
    "'FIX_AXIS_PAR_LIST'",
    "'MAX_GRAD'",
    "'COMPARISON_QUANTITY'",
    "'DEPENDENT_CHARACTERISTIC'",
    "'MAP_LIST'",
    "'NUMBER'",
    "'VIRTUAL_CHARACTERISTIC'",
    "'COMPU_METHOD'",
    "'IDENTICAL'",
    "'FORM'",
    "'LINEAR'",
    "'RAT_FUNC'",
    "'TAB_INTP'",
    "'TAB_NOINTP'",
    "'TAB_VERB'",
    "'COEFFS'",
    "'COEFFS_LINEAR'",
    "'COMPU_TAB_REF'",
    "'FORMULA'",
    "'FORMULA_INV'",
    "'STATUS_STRING_REF'",
    "'COMPU_TAB'",
    "'DEFAULT_VALUE_NUMERIC'",
    "'COMPU_VTAB'",
    "'COMPU_VTAB_RANGE'",
    "'FRAME'",
    "'FRAME_MEASUREMENT'",
    "'FUNCTION'",
    "'DEF_CHARACTERISTIC'",
    "'FUNCTION_VERSION'",
    "'IN_MEASUREMENT'",
    "'LOC_MEASUREMENT'",
    "'OUT_MEASUREMENT'",
    "'SUB_FUNCTION'",
    "'GROUP'",
    "'REF_MEASUREMENT'",
    "'ROOT'",
    "'SUB_GROUP'",
    "'INSTANCE'",
    "'MEASUREMENT'",
    "'ARRAY_SIZE'",
    "'BIT_OPERATION'",
    "'LEFT_SHIFT'",
    "'RIGHT_SHIFT'",
    "'SIGN_EXTEND'",
    "'ECU_ADDRESS'",
    "'ERROR_MASK'",
    "'LAYOUT'",
    "'ROW_DIR'",
    "'COLUMN_DIR'",
    "'READ_WRITE'",
    "'VIRTUAL'",
    "'MOD_COMMON'",
    "'DATA_SIZE'",
    "'S_REC_LAYOUT'",
    "'MOD_PAR'",
    "'ADDR_EPK'",
    "'CALIBRATION_METHOD'",
    "'CALIBRATION_HANDLE'",
    "'CALIBRATION_HANDLE_TEXT'",
    "'CPU_TYPE'",
    "'CUSTOMER'",
    "'CUSTOMER_NO'",
    "'ECU'",
    "'ECU_CALIBRATION_OFFSET'",
    "'EPK'",
    "'MEMORY_LAYOUT'",
    "'PRG_CODE'",
    "'PRG_DATA'",
    "'PRG_RESERVED'",
    "'MEMORY_SEGMENT'",
    "'CALIBRATION_VARIABLES'",
    "'CODE'",
    "'DATA'",
    "'EXCLUDE_FROM_FLASH'",
    "'OFFLINE_DATA'",
    "'RESERVED'",
    "'SERAM'",
    "'VARIABLES'",
    "'EEPROM'",
    "'EPROM'",
    "'FLASH'",
    "'RAM'",
    "'ROM'",
    "'REGISTER'",
    "'INTERN'",
    "'EXTERN'",
    "'NO_OF_INTERFACES'",
    "'PHONE_NO'",
    "'SUPPLIER'",
    "'SYSTEM_CONSTANT'",
    "'USER'",
    "'RECORD_LAYOUT'",
    "'AXIS_PTS_X'",
    "'AXIS_PTS_Y'",
    "'AXIS_PTS_Z'",
    "'AXIS_PTS_4'",
    "'AXIS_PTS_5'",
    "'AXIS_RESCALE_X'",
    "'AXIS_RESCALE_Y'",
    "'AXIS_RESCALE_Z'",
    "'AXIS_RESCALE_4'",
    "'AXIS_RESCALE_5'",
    "'DIST_OP_X'",
    "'DIST_OP_Y'",
    "'DIST_OP_Z'",
    "'DIST_OP_4'",
    "'DIST_OP_5'",
    "'FIX_NO_AXIS_PTS_X'",
    "'FIX_NO_AXIS_PTS_Y'",
    "'FIX_NO_AXIS_PTS_Z'",
    "'FIX_NO_AXIS_PTS_4'",
    "'FIX_NO_AXIS_PTS_5'",
    "'FNC_VALUES'",
    "'ALTERNATE_CURVES'",
    "'ALTERNATE_WITH_X'",
    "'ALTERNATE_WITH_Y'",
    "'IDENTIFICATION'",
    "'NO_AXIS_PTS_X'",
    "'NO_AXIS_PTS_Y'",
    "'NO_AXIS_PTS_Z'",
    "'NO_AXIS_PTS_4'",
    "'NO_AXIS_PTS_5'",
    "'STATIC_RECORD_LAYOUT'",
    "'NO_RESCALE_X'",
    "'NO_RESCALE_Y'",
    "'NO_RESCALE_Z'",
    "'NO_RESCALE_4'",
    "'NO_RESCALE_5'",
    "'OFFSET_X'",
    "'OFFSET_Y'",
    "'OFFSET_Z'",
    "'OFFSET_4'",
    "'OFFSET_5'",
    "'RIP_ADDR_W'",
    "'RIP_ADDR_X'",
    "'RIP_ADDR_Y'",
    "'RIP_ADDR_Z'",
    "'RIP_ADDR_4'",
    "'RIP_ADDR_5'",
    "'SHIFT_OP_X'",
    "'SHIFT_OP_Y'",
    "'SHIFT_OP_Z'",
    "'SHIFT_OP_4'",
    "'SHIFT_OP_5'",
    "'SRC_ADDR_X'",
    "'SRC_ADDR_Y'",
    "'SRC_ADDR_Z'",
    "'SRC_ADDR_4'",
    "'SRC_ADDR_5'",
    "'TYPEDEF_CHARACTERISTIC'",
    "'TYPEDEF_MEASUREMENT'",
    "'TYPEDEF_STRUCTURE'",
    "'STRUCTURE_COMPONENT'",
    "'UNIT'",
    "'DERIVED'",
    "'EXTENDED_SI'",
    "'SI_EXPONENTS'",
    "'UNIT_CONVERSION'",
    "'USER_RIGHTS'",
    "'REF_GROUP'",
    "'VARIANT_CODING'",
    "'VAR_CHARACTERISTIC'",
    "'VAR_ADDRESS'",
    "'VAR_CRITERION'",
    "'VAR_MEASUREMENT'",
    "'VAR_SELECTION_CHARACTERISTIC'",
    "'VAR_FORBIDDEN_COMB'",
    "'VAR_NAMING'",
    "'NUMERIC'",
    "'APLHA'",
    "'VAR_SEPARATOR'",
    "'.'",
    "'['",
    "']'",
    "'UBYTE'",
    "'SBYTE'",
    "'UWORD'",
    "'SWORD'",
    "'ULONG'",
    "'SLONG'",
    "'A_UINT64'",
    "'A_INT64'",
    "'FLOAT16_IEEE'",
    "'FLOAT32_IEEE'",
    "'FLOAT64_IEEE'",
    "'BYTE'",
    "'WORD'",
    "'LONG'",
    "'PBYTE'",
    "'PWORD'",
    "'PLONG'",
    "'DIRECT'",
    "'LITTLE_ENDIAN'",
    "'BIG_ENDIAN'",
    "'MSB_LAST'",
    "'MSB_FIRST'",
    "'INDEX_INCR'",
    "'INDEX_DECR'",
    "'SYMBOL_TYPE_LINK'",
    "'/begin'",
    "'/end'",
]


############################
############################
############################
MAY_KEYWORDS = {
    "ABSOLUTE": [("DEPOSIT", 1)],
    "ALTERNATE_CURVES": [("FNC_VALUES", 3)],
    "ALTERNATE_WITH_X": [("FNC_VALUES", 3)],
    "ALTERNATE_WITH_Y": [("FNC_VALUES", 3)],
    "APLHA": [("VAR_NAMING", 1)],
    "ASCII": [("CHARACTERISTIC", 3), ("TYPEDEF_CHARACTERISTIC", 3)],
    "CALIBRATION": [("CALIBRATION_ACCESS", 1)],
    "CALIBRATION_VARIABLES": [("MEMORY_SEGMENT", 3)],
    "CODE": [("MEMORY_SEGMENT", 3)],
    "COLUMN_DIR": [("FNC_VALUES", 3), ("LAYOUT", 1)],
    "COM_AXIS": [("AXIS_DESCR", 1)],
    "CUBE_4": [("CHARACTERISTIC", 3), ("TYPEDEF_CHARACTERISTIC", 3)],
    "CUBE_5": [("CHARACTERISTIC", 3), ("TYPEDEF_CHARACTERISTIC", 3)],
    "CUBOID": [("CHARACTERISTIC", 3), ("TYPEDEF_CHARACTERISTIC", 3)],
    "CURVE": [("CHARACTERISTIC", 3), ("TYPEDEF_CHARACTERISTIC", 3)],
    "CURVE_AXIS": [("AXIS_DESCR", 1)],
    "DATA": [("MEMORY_SEGMENT", 3)],
    "DERIVED": [("UNIT", 4)],
    "DIFFERENCE": [("DEPOSIT", 1)],
    "EEPROM": [("MEMORY_SEGMENT", 4)],
    "EPROM": [("MEMORY_SEGMENT", 4)],
    "EXCLUDE_FROM_FLASH": [("MEMORY_SEGMENT", 3)],
    "EXTENDED_SI": [("UNIT", 4)],
    "EXTERN": [("MEMORY_SEGMENT", 5)],
    "FIX_AXIS": [("AXIS_DESCR", 1)],
    "FLASH": [("MEMORY_SEGMENT", 4)],
    "FORM": [("COMPU_METHOD", 3)],
    "IDENTICAL": [("COMPU_METHOD", 3)],
    "INTERN": [("MEMORY_SEGMENT", 5)],
    "LINEAR": [("COMPU_METHOD", 3)],
    "MAP": [("CHARACTERISTIC", 3), ("TYPEDEF_CHARACTERISTIC", 3)],
    "MONOTONOUS": [("MONOTONY", 1)],
    "MON_DECREASE": [("MONOTONY", 1)],
    "MON_INCREASE": [("MONOTONY", 1)],
    "NOT_IN_MCD_SYSTEM": [("CALIBRATION_ACCESS", 1)],
    "NOT_MON": [("MONOTONY", 1)],
    "NO_CALIBRATION": [("CALIBRATION_ACCESS", 1)],
    "NUMERIC": [("VAR_NAMING", 1)],
    "OFFLINE_CALIBRATION": [("CALIBRATION_ACCESS", 1)],
    "OFFLINE_DATA": [("MEMORY_SEGMENT", 3)],
    "PRG_CODE": [("MEMORY_LAYOUT", 1)],
    "PRG_DATA": [("MEMORY_LAYOUT", 1)],
    "PRG_RESERVED": [("MEMORY_LAYOUT", 1)],
    "RAM": [("MEMORY_SEGMENT", 4)],
    "RAT_FUNC": [("COMPU_METHOD", 3)],
    "REGISTER": [("MEMORY_SEGMENT", 4)],
    "RESERVED": [("MEMORY_SEGMENT", 3)],
    "RES_AXIS": [("AXIS_DESCR", 1)],
    "ROM": [("MEMORY_SEGMENT", 4)],
    "ROW_DIR": [("FNC_VALUES", 3), ("LAYOUT", 1)],
    "SERAM": [("MEMORY_SEGMENT", 3)],
    "STD_AXIS": [("AXIS_DESCR", 1)],
    "STRICT_DECREASE": [("MONOTONY", 1)],
    "STRICT_INCREASE": [("MONOTONY", 1)],
    "STRICT_MON": [("MONOTONY", 1)],
    "TAB_INTP": [("COMPU_METHOD", 3), ("COMPU_TAB", 3)],
    "TAB_NOINTP": [("COMPU_METHOD", 3), ("COMPU_TAB", 3)],
    "TAB_VERB": [("COMPU_METHOD", 3), ("COMPU_VTAB", 3)],
    "VALUE": [("CHARACTERISTIC", 3), ("TYPEDEF_CHARACTERISTIC", 3)],
    "VAL_BLK": [("CHARACTERISTIC", 3), ("TYPEDEF_CHARACTERISTIC", 3)],
    "VARIABLES": [("MEMORY_SEGMENT", 3)],
}
############################
############################
############################

HEADER = """/*
**
**  !!! AUTOMATICALLY GENERATED FILE -- DO NOT EDIT !!!
**
*/
#if !defined(__TOKEN_TYPE_HPP)
#define __TOKEN_TYPE_HPP

#include <map>
#include <string>
#include <cstdint>

"""

FOOTER = """
#endif  // __TOKEN_TYPE_HPP
"""

HEADER2 = """/*
**
**  !!! AUTOMATICALLY GENERATED FILE -- DO NOT EDIT !!!
**
*/
#if !defined(__A2LTOKEN_HPP)
#define __A2LTOKEN_HPP

#include <map>
#include <string>
#include <cstdint>

"""

FOOTER2 = """
#endif  // __A2LTOKEN_HPP
"""

TPs = (
    ("INVALID", 0),
    ("BEGIN", 276),
    ("END", 277),
    ("IDENT", 278),
    ("FLOAT", 279),
    ("INT", 280),
    ("HEX", 281),
    ("COMMENT", 282),
    ("WS", 283),
    ("STRING", 284),
)

with open("token_type.hpp", "wt") as of:
    of.write(HEADER)
    of.write("const std::map<std::string, int> A2L_KEYWORDS {\n")
    for idx, item in enumerate(literalNames):
        item = item.replace("'", '"')
        if not item.startswith('"'):
            item = f'"{item}"'
        of.write(f"""    {{{item}, {idx}}},\n""")
    of.write("};\n\n")
    of.write("enum class TokenType: uint16_t {\n")
    for idx in range(275):
        of.write(f"\tT__{idx}={idx + 1},\n")
    for n, v in TPs:
        of.write(f"\t{n}={v},\n")
    of.write("};\n")
    of.write(FOOTER)

with open("a2ltoken.hpp", "wt") as of:
    of.write(HEADER2)
    of.write("enum class A2LTokenType: uint16_t {\n")

    li = [(idx, item) for idx, item in enumerate(literalNames)]
    li.extend([(v, k) for k, v in TPs])
    for idx, item in sorted(li):
        item = item.replace("'", "")
        if item in ("/begin", "/end", "<INVALID>"):
            continue
        if item == ".":
            item = "DOT"
        elif item == "[":
            item = "BRO"
        elif item == "]":  #
            item = "BRC"

        # if not item.startswith('"'):
        #    item = f'"{item}"'
        of.write(f"""    {item} = {idx},\n""")
    of.write("};\n\n")
    of.write(FOOTER2)


with open("a2l.tokens", "wt") as of:
    for idx in range(275):
        of.write(f"T__{idx}={idx + 1}\n")
    for n, v in TPs:
        if v == 0:
            continue
        of.write(f"{n}={v}\n")
    for idx, item in enumerate(literalNames):
        of.write(f"""{item}={idx}\n""")



#!/usr/bin/env python

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2024 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.1.0"

import logging


try:
    from rich.logging import RichHandler

    RICH_AVAILABLE = True
except ImportError:
    RICH_AVAILABLE = False


class Logger:

    LOGGER_BASE_NAME = "pya2l"
    FORMAT = "[%(levelname)s (%(name)s)]: %(message)s"

    def __init__(self, name, level="INFO"):
        if not logging.getLogger().handlers:
            self.logger = logging.getLogger(f"{self.LOGGER_BASE_NAME}.{name}")
            # self.logger.setLevel(level.upper())
            self.setLevel(level)
            if RICH_AVAILABLE:
                handler = RichHandler()
            else:
                handler = logging.StreamHandler()
            # handler.setLevel(level.upper())
            formatter = logging.Formatter(self.FORMAT)
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)
        if RICH_AVAILABLE:
            self.logger = logging.getLogger("rich")
        else:
            self.logger = logging.getLogger(f"{self.LOGGER_BASE_NAME}.{name}")
        self.lastMessage = None
        self.lastSeverity = None

    def getLastError(self):
        result = (self.lastSeverity, self.lastMessage)
        self.lastSeverity = self.lastMessage = None
        return result

    def log(self, message, level):
        self.lastSeverity = level
        self.lastMessage = message
        self.logger.log(level, f"{message}")
        # print("{0}{1}".format(level, message))

    def info(self, message):
        self.log(message, logging.INFO)

    def warn(self, message):
        self.log(message, logging.WARN)

    def debug(self, message):
        self.log(message, logging.DEBUG)

    def error(self, message):
        self.log(message, logging.ERROR)

    def critical(self, message):
        self.log(message, logging.CRITICAL)

    def verbose(self):
        self.logger.setLevel(logging.DEBUG)

    def silent(self):
        self.logger.setLevel(logging.CRITICAL)

    def setLevel(self, level):
        LEVEL_MAP = {
            "INFO": logging.INFO,
            "WARN": logging.WARN,
            "DEBUG": logging.DEBUG,
            "ERROR": logging.ERROR,
            "CRITICAL": logging.CRITICAL,
        }
        if isinstance(level, str):
            level = LEVEL_MAP.get(level.upper(), logging.WARN)
        self.logger.setLevel(level)



import re
from dataclasses import dataclass
from enum import IntEnum
from typing import Any, List, Optional


class IfDataTokenType(IntEnum):
    NONE = 0
    IDENT = 1
    FLOAT = 2
    INT = 3
    COMMENT = 4
    STRING = 6
    BEGIN = 7
    END = 8
    WS = 9


@dataclass
class IfDataToken:
    type: IfDataTokenType
    value: Any


PAT_WS = re.compile(r"\s+")
PAT_BEGIN = re.compile("/begin")
PAT_END = re.compile("/end")

PAT_COMMENT = re.compile(r"(/\*.*?\*/)|(//.*?)")
PAT_STRING = re.compile(r'"[^"]*"')

PAT_FLOAT = re.compile(r"[+\-]?(\d+([.]\d*)?([eE][+\-]?\d+)?|[.]\d+([eE][+\-]?\d+)?)")

PAT_INT = re.compile(r"(0x[0-9a-fA-F]+)|([+\-]?[0-9]+)")
PAT_IDENT = re.compile(r"([a-zA-Z_][a-zA-Z_0-9.]*)\b")


def identity(value: Any) -> Any:
    return value


def stripper(text: str) -> str:
    return text.strip('"')


def convert_to_float(value: str) -> float:
    return float(value)


def convert_to_int(value: str) -> int:
    base = 16 if value.lower().startswith("0x") else 10
    return int(value, base)


EXPRESSIONS = [
    (IfDataTokenType.WS, PAT_WS, identity),
    (IfDataTokenType.COMMENT, PAT_COMMENT, identity),
    (IfDataTokenType.STRING, PAT_STRING, stripper),
    (IfDataTokenType.INT, PAT_INT, convert_to_int),
    (IfDataTokenType.FLOAT, PAT_FLOAT, convert_to_float),
    (IfDataTokenType.BEGIN, PAT_BEGIN, identity),
    (IfDataTokenType.END, PAT_END, identity),
    (IfDataTokenType.IDENT, PAT_IDENT, identity),
]


class IfDataLexer:

    def __init__(self, section: str, skip_list: list[IfDataTokenType] = [IfDataTokenType.WS, IfDataTokenType.COMMENT]):
        self.section = section
        self.section_length = len(section)
        self.skip_list = skip_list
        self.pos = 0

    def run(self) -> List[IfDataToken]:
        result = []
        while self.pos < self.section_length:
            token = self.get_token()
            if token is not None:
                result.append(token)
        return result

    def get_token(self) -> Optional[IfDataToken]:
        for code, expr, converter in EXPRESSIONS:
            match = expr.match(self.section, self.pos)
            if match:
                self.pos += match.end() - match.start()
                text = match.group()
                if code in self.skip_list:
                    return None
                return IfDataToken(code, converter(text))
        print(f"Invalid char {self.section[self.pos]!r}")
        self.pos += 1
        return None


def ifdata_lexer(
    section: str, skip_list: list[IfDataTokenType] = [IfDataTokenType.WS, IfDataTokenType.COMMENT]
) -> List[IfDataToken]:
    length = len(section)
    pos = 0
    result = []

    while pos < length:
        found = False
        for code, expr in EXPRESSIONS:
            match = expr.match(section, pos)
            if match:
                pos += match.end() - match.start()
                if pos > 60:
                    print()
                text = match.group()
                if code in skip_list:
                    continue
                print(code.name, text)
                found = True
                result.append(IfDataToken(code, text))
                continue
        if not found:
            print(f"Invalid char {section[pos]!r}")
            pos += 1
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest

from pya2l import exceptions
from pya2l import model
from pya2l.a2l_listener import A2LListener
from pya2l.api.inspect import CompuMethod
from pya2l.api.inspect import Group
from pya2l.api.inspect import Measurement
from pya2l.api.inspect import ModCommon
from pya2l.api.inspect import ModPar
from pya2l.api.inspect import NoCompuMethod
from pya2l.api.inspect import TypedefStructure
from pya2l.parserlib import ParserWrapper


try:
    import numpy as np
except ImportError:
    has_numpy = False
else:
    has_numpy = True

try:
    from scipy.interpolate import RegularGridInterpolator
except ImportError:
    has_scipy = False
else:
    has_scipy = True


RUN_MATH_TEST = has_numpy and has_scipy


def test_measurement_basic():
    parser = ParserWrapper("a2l", "measurement", A2LListener, debug=False)
    DATA = """
    /begin MEASUREMENT
        N /* name */
        "Engine speed" /* long identifier */
        UWORD /* datatype */
        NO_COMPU_METHOD /* conversion */
        2 /* resolution */
        2.5 /* accuracy */
        120.0 /* lower limit */
        8400.0 /* upper limit */
    /end MEASUREMENT
    """
    db, _ = parser.parseFromString(DATA)
    meas = Measurement(db.session, "N")
    assert meas.name == "N"
    assert meas.longIdentifier == "Engine speed"
    assert meas.datatype == "UWORD"
    assert meas.resolution == 2
    assert meas.accuracy == 2.5
    assert meas.lowerLimit == 120
    assert meas.upperLimit == 8400.0
    assert meas.annotations == []
    assert meas.arraySize is None
    assert meas.bitMask is None
    assert meas.bitOperation is None
    assert meas.byteOrder is None
    assert meas.discrete is False
    assert meas.displayIdentifier is None
    assert meas.ecuAddress is None
    # assert meas.ecuAddressExtension is None
    assert meas.errorMask is None
    assert meas.format is None
    assert meas.functionList == []
    assert meas.layout is None
    assert meas.matrixDim is None
    assert meas.maxRefresh is None
    assert meas.physUnit is None
    assert meas.readWrite is False
    assert meas.refMemorySegment is None
    assert meas.symbolLink is None
    assert meas.virtual == []
    assert meas.compuMethod is NoCompuMethod()


def test_measurement_full_featured():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD Velocity
            "conversion method for velocity"
            RAT_FUNC
            "%6.2"
            "km/h"
            COEFFS 0 100 0 0 0 1
        /end COMPU_METHOD

        /begin MEASUREMENT
            N /* name */
            "Engine speed" /* long identifier */
            UWORD /* datatype */
            Velocity /* conversion */
            2 /* resolution */
            2.5 /* accuracy */
            120.0 /* lower limit */
            8400.0 /* upper limit */
            /begin ANNOTATION
                ANNOTATION_LABEL "ASAM Workinggroup"
                ANNOTATION_ORIGIN "Python Universe"
                /begin ANNOTATION_TEXT
                    "Test the A2L annotation"
                    "Another line"
                /end ANNOTATION_TEXT
            /end ANNOTATION
            ARRAY_SIZE 8 /* array of 8 values */
            BIT_MASK 0x0FFF
            BYTE_ORDER MSB_FIRST
            /begin BIT_OPERATION
                RIGHT_SHIFT 4 /*4 positions*/
                SIGN_EXTEND
            /end BIT_OPERATION
            BYTE_ORDER MSB_FIRST
            DISCRETE
            DISPLAY_IDENTIFIER load_engine
            ECU_ADDRESS 0xcafebabe
            ECU_ADDRESS_EXTENSION 42
            ERROR_MASK 0x00000001
            FORMAT "%4.2f"
            /begin FUNCTION_LIST
                ID_ADJUSTM
                FL_ADJUSTM
            /end FUNCTION_LIST
            LAYOUT ROW_DIR
            MATRIX_DIM  2   4   3
            MAX_REFRESH 3 15 /* 15 msec */
            PHYS_UNIT "mph"
            READ_WRITE
            REF_MEMORY_SEGMENT Data2
            SYMBOL_LINK "VehicleSpeed" /* Symbol name */
                    4711
            /begin VIRTUAL
                PHI_BASIS
                PHI_CORR
            /end VIRTUAL
        /end MEASUREMENT
    /end MODULE
    """

    db, _ = parser.parseFromString(DATA)
    meas = Measurement(db.session, "N")
    assert meas.name == "N"
    assert meas.longIdentifier == "Engine speed"
    assert meas.datatype == "UWORD"
    assert meas.resolution == 2
    assert meas.accuracy == 2.5
    assert meas.lowerLimit == 120
    assert meas.upperLimit == 8400.0
    assert meas.arraySize == 8
    assert meas.bitMask == 0x0FFF
    assert meas.annotations == [
        {
            "label": "ASAM Workinggroup",
            "origin": "Python Universe",
            "text": ["Test the A2L annotation", "Another line"],
        }
    ]
    assert meas.bitOperation == {"amount": 4, "direction": "R", "sign_extend": True}
    assert meas.byteOrder == "MSB_FIRST"
    assert meas.discrete
    assert meas.displayIdentifier == "load_engine"
    assert meas.ecuAddress == 0xCAFEBABE
    assert meas.ecuAddressExtension == 42
    assert meas.errorMask == 1
    assert meas.format == "%4.2f"
    assert meas.functionList == ["ID_ADJUSTM", "FL_ADJUSTM"]
    assert meas.layout == "ROW_DIR"
    assert meas.matrixDim == {"x": 2, "y": 4, "z": 3}
    assert meas.maxRefresh == {"rate": 15, "scalingUnit": 3}
    assert meas.physUnit == "mph"
    assert meas.readWrite
    assert meas.refMemorySegment == "Data2"
    assert meas.symbolLink == {"offset": 4711, "symbolName": "VehicleSpeed"}
    assert meas.virtual == ["PHI_BASIS", "PHI_CORR"]
    assert meas.compuMethod.conversionType == "RAT_FUNC"
    assert meas.compuMethod.unit == "km/h"
    assert meas.compuMethod.format == "%6.2"
    assert meas.compuMethod.longIdentifier == "conversion method for velocity"
    assert meas.compuMethod.coeffs["a"] == 0.0
    assert meas.compuMethod.coeffs["b"] == 100.0
    assert meas.compuMethod.coeffs["c"] == 0.0
    assert meas.compuMethod.coeffs["d"] == 0.0
    assert meas.compuMethod.coeffs["e"] == 0.0
    assert meas.compuMethod.coeffs["f"] == 1.0


def test_measurement_no_compu_method():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin MEASUREMENT
            N /* name */
            "Engine speed" /* long identifier */
            UWORD /* datatype */
            NO_COMPU_METHOD /* conversion */
            2 /* resolution */
            2.5 /* accuracy */
            120.0 /* lower limit */
            8400.0 /* upper limit */
        /end MEASUREMENT
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    meas = Measurement(db.session, "N")
    assert meas.compuMethod is NoCompuMethod()


def test_measurement_compu_method_identical():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.IDENTICAL
            "conversion that delivers always phys = int"
            IDENTICAL "%3.0" "hours"
        /end COMPU_METHOD
        /begin MEASUREMENT
            N /* name */
            "Engine speed" /* long identifier */
            UWORD /* datatype */
            CM.IDENTICAL /* conversion */
            2 /* resolution */
            2.5 /* accuracy */
            120.0 /* lower limit */
            8400.0 /* upper limit */
        /end MEASUREMENT
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    meas = Measurement(db.session, "N")
    assert meas.compuMethod.format == "%3.0"
    assert meas.compuMethod.conversionType == "IDENTICAL"
    assert meas.compuMethod.unit == "hours"
    assert meas.compuMethod.longIdentifier == "conversion that delivers always phys = int"


def test_measurement_compu_method_form():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.FORM.X_PLUS_4
          "" FORM "%6.1" "rpm"
          /begin FORMULA
            "X1+4"
            FORMULA_INV "X1-4"
            /end FORMULA
        /end COMPU_METHOD
        /begin MEASUREMENT
            N /* name */
            "Engine speed" /* long identifier */
            UWORD /* datatype */
            CM.FORM.X_PLUS_4 /* conversion */
            2 /* resolution */
            2.5 /* accuracy */
            120.0 /* lower limit */
            8400.0 /* upper limit */
        /end MEASUREMENT
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    meas = Measurement(db.session, "N")
    assert meas.compuMethod.format == "%6.1"
    assert meas.compuMethod.conversionType == "FORM"
    assert meas.compuMethod.unit == "rpm"
    assert meas.compuMethod.longIdentifier == ""
    assert meas.compuMethod.formula["formula"] == "X1+4"
    assert meas.compuMethod.formula["formula_inv"] == "X1-4"


def test_measurement_compu_method_linear():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.LINEAR.MUL_2
          "Linear function with parameter set for phys = f(int) = 2*int + 0"
          LINEAR "%3.1" "m/s"
          COEFFS_LINEAR 2 0
        /end COMPU_METHOD
        /begin MEASUREMENT
            N /* name */
            "Engine speed" /* long identifier */
            UWORD /* datatype */
            CM.LINEAR.MUL_2 /* conversion */
            2 /* resolution */
            2.5 /* accuracy */
            120.0 /* lower limit */
            8400.0 /* upper limit */
        /end MEASUREMENT
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    meas = Measurement(db.session, "N")

    assert meas.compuMethod.format == "%3.1"
    assert meas.compuMethod.conversionType == "LINEAR"
    assert meas.compuMethod.unit == "m/s"
    assert meas.compuMethod.longIdentifier == "Linear function with parameter set for phys = f(int) = 2*int + 0"
    assert meas.compuMethod.coeffs_linear["a"] == 2.0
    assert meas.compuMethod.coeffs_linear["b"] == 0.0


def test_measurement_compu_method_rat_func():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.RAT_FUNC.DIV_81_9175
          "rational function with parameter set for impl = f(phys) = phys * 81.9175"
          RAT_FUNC "%8.4" "grad C"
          COEFFS 0 81.9175 0 0 0 1
        /end COMPU_METHOD
        /begin MEASUREMENT
            N /* name */
            "Engine speed" /* long identifier */
            UWORD /* datatype */
            CM.RAT_FUNC.DIV_81_9175 /* conversion */
            2 /* resolution */
            2.5 /* accuracy */
            120.0 /* lower limit */
            8400.0 /* upper limit */
        /end MEASUREMENT
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    meas = Measurement(db.session, "N")

    assert meas.compuMethod.format == "%8.4"
    assert meas.compuMethod.conversionType == "RAT_FUNC"
    assert meas.compuMethod.unit == "grad C"
    assert meas.compuMethod.longIdentifier == "rational function with parameter set for impl = f(phys) = phys * 81.9175"
    assert meas.compuMethod.coeffs["a"] == 0.0
    assert meas.compuMethod.coeffs["b"] == 81.9175
    assert meas.compuMethod.coeffs["c"] == 0.0
    assert meas.compuMethod.coeffs["d"] == 0.0
    assert meas.compuMethod.coeffs["e"] == 0.0
    assert meas.compuMethod.coeffs["f"] == 1.0


def test_measurement_compu_method_tab_intp():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.TAB_INTP.DEFAULT_VALUE
          ""
          TAB_INTP "%8.4" "U/min"
          COMPU_TAB_REF CM.TAB_INTP.DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_TAB CM.TAB_INTP.DEFAULT_VALUE.REF
           ""
           TAB_INTP
           12
           -3 98
           -1 99
           0 100
           2 102
           4 104
           5 105
           6 106
           7 107
           8 108
           9 109
           10 110
           13 111
           DEFAULT_VALUE_NUMERIC 300.56
        /end COMPU_TAB
        /begin MEASUREMENT
            N /* name */
            "Engine speed" /* long identifier */
            UWORD /* datatype */
            CM.TAB_INTP.DEFAULT_VALUE /* conversion */
            2 /* resolution */
            2.5 /* accuracy */
            120.0 /* lower limit */
            8400.0 /* upper limit */
        /end MEASUREMENT
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    meas = Measurement(db.session, "N")
    assert meas.compuMethod.format == "%8.4"
    assert meas.compuMethod.conversionType == "TAB_INTP"
    assert meas.compuMethod.unit == "U/min"
    assert meas.compuMethod.longIdentifier == ""
    assert meas.compuMethod.tab["default_value"] == 300.56
    assert meas.compuMethod.tab["interpolation"]
    assert meas.compuMethod.tab["num_values"] == 12
    assert meas.compuMethod.tab["in_values"] == [
        -3.0,
        -1.0,
        0.0,
        2.0,
        4.0,
        5.0,
        6.0,
        7.0,
        8.0,
        9.0,
        10.0,
        13.0,
    ]
    assert meas.compuMethod.tab["out_values"] == [
        98.0,
        99.0,
        100.0,
        102.0,
        104.0,
        105.0,
        106.0,
        107.0,
        108.0,
        109.0,
        110.0,
        111.0,
    ]


def test_measurement_compu_method_tab_verb():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.TAB_VERB.DEFAULT_VALUE
          "Verbal conversion with default value"
          TAB_VERB "%12.0" ""
          COMPU_TAB_REF CM.TAB_VERB.DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_VTAB CM.TAB_VERB.DEFAULT_VALUE.REF
          "List of text strings and relation to impl value"
          TAB_VERB 6
          2 "red"
          3 "orange"
          4 "yellow"
          5 "green"
          6 "blue"
          7 "violet"
          DEFAULT_VALUE "unknown signal type"
        /end COMPU_VTAB
        /begin MEASUREMENT
            N /* name */
            "Engine speed" /* long identifier */
            UWORD /* datatype */
            CM.TAB_VERB.DEFAULT_VALUE /* conversion */
            2 /* resolution */
            2.5 /* accuracy */
            120.0 /* lower limit */
            8400.0 /* upper limit */
        /end MEASUREMENT
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    meas = Measurement(db.session, "N")
    assert meas.compuMethod.format == "%12.0"
    assert meas.compuMethod.conversionType == "TAB_VERB"
    assert meas.compuMethod.unit == ""
    assert meas.compuMethod.longIdentifier == "Verbal conversion with default value"
    assert meas.compuMethod.tab_verb["default_value"] == "unknown signal type"
    assert meas.compuMethod.tab_verb["num_values"] == 6
    assert meas.compuMethod.tab_verb["ranges"] is False
    assert meas.compuMethod.tab_verb["in_values"] == [2.0, 3.0, 4.0, 5.0, 6.0, 7.0]
    assert meas.compuMethod.tab_verb["text_values"] == [
        "red",
        "orange",
        "yellow",
        "green",
        "blue",
        "violet",
    ]


def test_measurement_compu_method_tab_verb_range():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.VTAB_RANGE.DEFAULT_VALUE
           "verbal range with default value"
           TAB_VERB
           "%4.2"
           ""
           COMPU_TAB_REF CM.VTAB_RANGE.DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_VTAB_RANGE CM.VTAB_RANGE.DEFAULT_VALUE.REF
           ""
           11
           0 1 "Zero_to_one"
           2 3 "two_to_three"
           4 7 "four_to_seven"
           14 17 "fourteen_to_seventeen"
           18 99 "eigteen_to_ninetynine"
           100 100 "hundred"
           101 101 "hundredone"
           102 102 "hundredtwo"
           103 103 "hundredthree"
           104 104 "hundredfour"
           105 105 "hundredfive"
           DEFAULT_VALUE "out of range value"
        /end COMPU_VTAB_RANGE
        /begin MEASUREMENT
            N /* name */
            "Engine speed" /* long identifier */
            UWORD /* datatype */
            CM.VTAB_RANGE.DEFAULT_VALUE /* conversion */
            2 /* resolution */
            2.5 /* accuracy */
            120.0 /* lower limit */
            8400.0 /* upper limit */
        /end MEASUREMENT
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    meas = Measurement.get(db.session, "N")
    assert meas.compuMethod.format == "%4.2"
    assert meas.compuMethod.conversionType == "TAB_VERB"
    assert meas.compuMethod.unit == ""
    assert meas.compuMethod.longIdentifier == "verbal range with default value"
    assert meas.compuMethod.name == "CM.VTAB_RANGE.DEFAULT_VALUE"
    assert meas.compuMethod.statusStringRef is None
    assert meas.compuMethod.tab_verb["default_value"] == "out of range value"
    assert meas.compuMethod.tab_verb["num_values"] == 11
    assert meas.compuMethod.tab_verb["ranges"]
    assert meas.compuMethod.tab_verb["lower_values"] == [
        0.0,
        2.0,
        4.0,
        14.0,
        18.0,
        100.0,
        101.0,
        102.0,
        103.0,
        104.0,
        105.0,
    ]
    assert meas.compuMethod.tab_verb["text_values"] == [
        "Zero_to_one",
        "two_to_three",
        "four_to_seven",
        "fourteen_to_seventeen",
        "eigteen_to_ninetynine",
        "hundred",
        "hundredone",
        "hundredtwo",
        "hundredthree",
        "hundredfour",
        "hundredfive",
    ]
    assert meas.compuMethod.tab_verb["upper_values"] == [
        1.0,
        3.0,
        7.0,
        17.0,
        99.0,
        100.0,
        101.0,
        102.0,
        103.0,
        104.0,
        105.0,
    ]


def test_mod_par_full_featured():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin MOD_PAR "Note: Provisional release for test purposes only!"
            VERSION "Test version of 01.02.1994"
            ADDR_EPK 0x45678
            EPK     "EPROM identifier test"
            SUPPLIER "M&K GmbH Chemnitz"
            CUSTOMER "LANZ-Landmaschinen"
            CUSTOMER_NO "0123456789"
            USER "A.N.Wender"
            PHONE_NO "09951 56456"
            ECU "Engine control"
            CPU_TYPE "Motorola 0815"
            NO_OF_INTERFACES 2
            /begin MEMORY_SEGMENT ext_Ram
                "external RAM"
                DATA
                RAM
                EXTERN
                0x30000
                0x1000
                -1 -1 -1 -1 -1
            /end MEMORY_SEGMENT
            /begin MEMORY_LAYOUT PRG_RESERVED
                0x0000
                0x0400
                -1 -1 -1 -1 -1
            /end MEMORY_LAYOUT
            /begin MEMORY_LAYOUT PRG_CODE
                0x0400
                0x3C00
                -1 -1 -1 -1 -1
            /end MEMORY_LAYOUT
            /begin MEMORY_LAYOUT PRG_DATA
                0x4000
                0x5800
                -1 -1 -1 -1 -1
            /end MEMORY_LAYOUT
            SYSTEM_CONSTANT "CONTROLLERx constant1" "0.33"
            SYSTEM_CONSTANT "CONTROLLERx constant2" "2.79"
        /end MOD_PAR
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    mp = ModPar(db.session, "testModule")
    assert mp.comment == "Note: Provisional release for test purposes only!"

    assert mp.version == "Test version of 01.02.1994"
    assert mp.addrEpk[0] == 284280
    assert mp.epk == "EPROM identifier test"
    assert mp.supplier == "M&K GmbH Chemnitz"
    assert mp.customer == "LANZ-Landmaschinen"
    assert mp.customerNo == "0123456789"
    assert mp.user == "A.N.Wender"
    assert mp.phoneNo == "09951 56456"
    assert mp.ecu == "Engine control"
    assert mp.cpu == "Motorola 0815"
    assert mp.noOfInterfaces == 2
    ms = mp.memorySegments[0]
    assert ms == {
        "address": 196608,
        "attribute": "EXTERN",
        "longIdentifier": "external RAM",
        "memoryType": "RAM",
        "name": "ext_Ram",
        "offset_0": -1,
        "offset_1": -1,
        "offset_2": -1,
        "offset_3": -1,
        "offset_4": -1,
        "prgType": "DATA",
        "size": 4096,
    }
    m0, m1, m2 = mp.memoryLayouts
    assert m0 == {
        "address": 0,
        "offset_0": -1,
        "offset_1": -1,
        "offset_2": -1,
        "offset_3": -1,
        "offset_4": -1,
        "prgType": "PRG_RESERVED",
        "size": 1024,
    }
    assert m1 == {
        "address": 1024,
        "offset_0": -1,
        "offset_1": -1,
        "offset_2": -1,
        "offset_3": -1,
        "offset_4": -1,
        "prgType": "PRG_CODE",
        "size": 15360,
    }
    assert m2 == {
        "address": 16384,
        "offset_0": -1,
        "offset_1": -1,
        "offset_2": -1,
        "offset_3": -1,
        "offset_4": -1,
        "prgType": "PRG_DATA",
        "size": 22528,
    }
    sc = mp.systemConstants
    assert sc == {"CONTROLLERx constant1": 0.33, "CONTROLLERx constant2": 2.79}


def test_mod_par_basic():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin MOD_PAR "Note: Provisional release for test purposes only!"
        /end MOD_PAR
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    mp = ModPar(db.session, "testModule")
    assert mp.comment == "Note: Provisional release for test purposes only!"
    assert mp.version is None
    assert mp.addrEpk == []
    assert mp.epk is None
    assert mp.supplier is None
    assert mp.customer is None
    assert mp.customerNo is None
    assert mp.user is None
    assert mp.phoneNo is None
    assert mp.ecu is None
    assert mp.cpu is None
    assert mp.noOfInterfaces is None
    assert mp.memorySegments == []
    assert mp.memoryLayouts == []
    assert mp.systemConstants == {}


def test_mod_common_full_featured():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin MOD_COMMON "Characteristic maps always deposited in same mode"
            S_REC_LAYOUT S_ABL
            DEPOSIT ABSOLUTE
            BYTE_ORDER MSB_LAST
            DATA_SIZE 16
            ALIGNMENT_BYTE  1
            ALIGNMENT_WORD          2
            ALIGNMENT_LONG          4
            ALIGNMENT_INT64         8
            ALIGNMENT_FLOAT32_IEEE  4
            ALIGNMENT_FLOAT64_IEEE  8

        /end MOD_COMMON
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    mc = ModCommon(db.session, "testModule")
    assert mc.comment == "Characteristic maps always deposited in same mode"
    # assert mc.sRecLayout.name == "S_ABL"
    assert mc.sRecLayout == "S_ABL"
    assert mc.deposit == "ABSOLUTE"
    assert mc.byteOrder == "MSB_LAST"
    assert mc.dataSize == 16
    assert mc.alignment == {
        "FLOAT64": 8,
        "DWORD": 4,
        "BYTE": 1,
        "WORD": 2,
        "QWORD": 8,
        "FLOAT32": 4,
        "FLOAT16": 2,
    }


@pytest.mark.skip
def test_group():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """
    /begin MODULE testModule ""
        /begin GROUP SOFTWARE_COMPONENTS
            "assignment of the definitions to C files"
            ROOT
            /begin SUB_GROUP INJE
                C6TD
            /end SUB_GROUP
        /end GROUP
        /begin GROUP INJE
            "Subsystem Injection"
            /begin SUB_GROUP Injec1
                Injec2
            /end SUB_GROUP
        /end GROUP
        /begin GROUP Injec1
            "Module filename Injec1"
            /begin REF_CHARACTERISTIC
                INJECTION_CURVE
            /end REF_CHARACTERISTIC
        /end GROUP
        /begin GROUP Injec2
            "Module filename Injec2"
            /begin REF_CHARACTERISTIC
                INJECTION_ADJUST
            /end REF_CHARACTERISTIC
            /begin REF_MEASUREMENT
                GAS_INPUT
                WHEEL_SPEED
            /end REF_MEASUREMENT
        /end GROUP
        /begin GROUP C6TD
            "Shift Point Control"
            /begin SUB_GROUP c6tdvder
                c6tdertf
            /end SUB_GROUP
        /end GROUP
        /begin GROUP c6tdvder
            "Module filename c6tdvder"
            /begin REF_CHARACTERISTIC
                SHIFT23_CURVE
            /end REF_CHARACTERISTIC
            /begin REF_MEASUREMENT
                LOOP_COUN2
                NO_GEAR
            /end REF_MEASUREMENT
        /end GROUP
        /begin GROUP c6tderft
            "Module filename c6tderft"
            /begin REF_CHARACTERISTIC
                LUP23_CURVE
            /end REF_CHARACTERISTIC
            /begin REF_MEASUREMENT
                TRANSMISSION_SP
                ENGINE_SPEED
            /end REF_MEASUREMENT
        /end GROUP
        /begin GROUP CALIBRATION_COMPONENTS
            "assignment of the definitions to
            calibration components"
            ROOT
            /begin SUB_GROUP
                Winter_Test
                Summer_Test
            /end SUB_GROUP
        /end GROUP
        /begin GROUP CALIBRATION_COMPONENTS_L4
            "L4-PCM 2002 cals"
            ROOT
            /begin SUB_GROUP LUFT
                CLOSED_LOOP
            /end SUB_GROUP
        /end GROUP
        /begin GROUP LUFT
        "Cals in LUFT Subsystem"
            /begin REF_CHARACTERISTIC
                KfLUFT_n_EngSpdThrsh
                KtLUFT_ScaledVE
                KaLUFT_AirPerCylCoeff
            /end REF_CHARACTERISTIC
        /end GROUP
        /begin GROUP CLOSED_LOOP
        "Cals in FCLS, FCLP & FCLL Subsystem"
            /begin REF_CHARACTERISTIC
                KaFCLP_U_O2LeanThrsh
                KfFCLP_t_O2AgainstMax
        /end REF_CHARACTERISTIC
        /end GROUP
        /begin GROUP Winter_Test
            "Flash this in winter time"
            /begin REF_CHARACTERISTIC
                GASOLINE_CURVE
            /end REF_CHARACTERISTIC
        /end GROUP
        /begin GROUP Summer_Test
            "Flash that in summer time"
            /begin REF_CHARACTERISTIC
                SUPER_CURVE
            /end REF_CHARACTERISTIC
        /end GROUP
        /begin GROUP SOFTWARE_COMPONENTS
            "L4-PCM 2002 C modules"
            ROOT
            /begin SUB_GROUP
                luftkmgr.c
                fclpkout.c
                viosmeng.c
            /end SUB_GROUP
        /end GROUP
        /begin GROUP luftkmgr.c
        "Objects in luftkmgr.c"
            /begin REF_CHARACTERISTIC
                KtLUFT_ScaledVE
            /end REF_CHARACTERISTIC
        /end GROUP
        /begin GROUP fclpkout.c
            "Objects in fclpkout.c"
            /begin REF_CHARACTERISTIC
                KaFCLP_U_O2LeanThrsh
                KfFCLP_t_O2AgainstMax
            /end REF_CHARACTERISTIC
        /end GROUP
        /begin GROUP viosmeng.c
            "Objects in viosmeng.c"
            /begin REF_CHARACTERISTIC
                VfVIOS_n_EngSpdLORES
                VfVIOS_p_AmbientAirPres
            /end REF_CHARACTERISTIC
        /end GROUP
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    gr = Group.get(db.session, "CALIBRATION_COMPONENTS")
    gr = Group.get(db.session, "CALIBRATION_COMPONENTS_L4")
    gr = Group.get(db.session, "SOFTWARE_COMPONENTS")
    print(gr)


def test_typedef_structure_no_instances():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """/begin MODULE testModule ""
        /begin RECORD_LAYOUT _UWORD FNC_VALUES 1 UWORD ROW_DIR DIRECT /end RECORD_LAYOUT
        /begin RECORD_LAYOUT _FLOAT64_IEEE FNC_VALUES 1 FLOAT64_IEEE ROW_DIR DIRECT /end RECORD_LAYOUT
        /begin RECORD_LAYOUT _UBYTE FNC_VALUES 1 UBYTE ROW_DIR DIRECT /end RECORD_LAYOUT
        /begin RECORD_LAYOUT _ULONG FNC_VALUES 1 ULONG ROW_DIR DIRECT /end RECORD_LAYOUT
        /begin TYPEDEF_STRUCTURE EcuTask "TYPEDEF for class EcuTask" 0x58 SYMBOL_TYPE_LINK "EcuTask"
            /begin STRUCTURE_COMPONENT taskId _UWORD 0x0 SYMBOL_TYPE_LINK "taskId" /end STRUCTURE_COMPONENT
            /begin STRUCTURE_COMPONENT counter _UWORD 0x2 SYMBOL_TYPE_LINK "counter" /end STRUCTURE_COMPONENT
            /begin STRUCTURE_COMPONENT channel1 _FLOAT64_IEEE 0x10 SYMBOL_TYPE_LINK "channel1" /end STRUCTURE_COMPONENT
            /begin STRUCTURE_COMPONENT byte _UBYTE 0x38 SYMBOL_TYPE_LINK "byte" /end STRUCTURE_COMPONENT
            /begin STRUCTURE_COMPONENT word _UWORD 0x3A SYMBOL_TYPE_LINK "word" /end STRUCTURE_COMPONENT
            /begin STRUCTURE_COMPONENT dword _ULONG 0x3C SYMBOL_TYPE_LINK "dword" /end STRUCTURE_COMPONENT
        /end TYPEDEF_STRUCTURE
    /end MODULE"""
    db, _ = parser.parseFromString(DATA)
    tdef = TypedefStructure.get(db.session, "EcuTask")
    assert tdef.name == "EcuTask"
    assert tdef.longIdentifier == "TYPEDEF for class EcuTask"
    assert tdef.size == 88
    assert tdef.link == "SYMBOL_TYPE_LINK"
    assert tdef.symbol == "EcuTask"
    assert tdef.instances == []
    components = tdef.components
    assert len(components) == 6
    comp = components[0]
    assert comp.name == "taskId"
    assert comp.offset == 0
    assert comp.link == "SYMBOL_TYPE_LINK"
    assert comp.symbol == "taskId"
    comp = components[1]
    assert comp.name == "counter"
    assert comp.offset == 2
    assert comp.link == "SYMBOL_TYPE_LINK"
    assert comp.symbol == "counter"
    comp = components[2]
    assert comp.name == "channel1"
    assert comp.offset == 16
    assert comp.link == "SYMBOL_TYPE_LINK"
    assert comp.symbol == "channel1"
    comp = components[3]
    assert comp.name == "byte"
    assert comp.offset == 0x38
    assert comp.link == "SYMBOL_TYPE_LINK"
    assert comp.symbol == "byte"
    comp = components[4]
    assert comp.name == "word"
    assert comp.offset == 0x3A
    assert comp.link == "SYMBOL_TYPE_LINK"
    assert comp.symbol == "word"
    comp = components[5]
    assert comp.name == "dword"
    assert comp.offset == 0x3C
    assert comp.link == "SYMBOL_TYPE_LINK"
    assert comp.symbol == "dword"


@pytest.mark.skip
def test_typedef_structure_with_instances():
    parser = ParserWrapper("a2l", "module", A2LListener, debug=False)
    DATA = """/begin MODULE testModule ""
        /begin RECORD_LAYOUT _UWORD FNC_VALUES 1 UWORD ROW_DIR DIRECT /end RECORD_LAYOUT
        /begin RECORD_LAYOUT _FLOAT64_IEEE FNC_VALUES 1 FLOAT64_IEEE ROW_DIR DIRECT /end RECORD_LAYOUT
        /begin RECORD_LAYOUT _UBYTE FNC_VALUES 1 UBYTE ROW_DIR DIRECT /end RECORD_LAYOUT
        /begin RECORD_LAYOUT _ULONG FNC_VALUES 1 ULONG ROW_DIR DIRECT /end RECORD_LAYOUT
        /begin TYPEDEF_STRUCTURE EcuTask "TYPEDEF for class EcuTask" 0x58 SYMBOL_TYPE_LINK "EcuTask"
            /begin STRUCTURE_COMPONENT taskId _UWORD 0x0 SYMBOL_TYPE_LINK "taskId" /end STRUCTURE_COMPONENT
            /begin STRUCTURE_COMPONENT counter _UWORD 0x2 SYMBOL_TYPE_LINK "counter" /end STRUCTURE_COMPONENT
            /begin STRUCTURE_COMPONENT channel1 _FLOAT64_IEEE 0x10 SYMBOL_TYPE_LINK "channel1" /end STRUCTURE_COMPONENT
            /begin STRUCTURE_COMPONENT byte _UBYTE 0x38 SYMBOL_TYPE_LINK "byte" /end STRUCTURE_COMPONENT
            /begin STRUCTURE_COMPONENT word _UWORD 0x3A SYMBOL_TYPE_LINK "word" /end STRUCTURE_COMPONENT
            /begin STRUCTURE_COMPONENT dword _ULONG 0x3C SYMBOL_TYPE_LINK "dword" /end STRUCTURE_COMPONENT
        /end TYPEDEF_STRUCTURE
        /begin INSTANCE ecuTask1 "ecupp task number 1" EcuTask 0x0 /end INSTANCE
        /begin INSTANCE ecuTask2 "ecu task number 2" EcuTask 0x0 /end INSTANCE
        /begin INSTANCE activeEcuTask "pointer to active ecu task" EcuTask 0x0 /end INSTANCE
    /end MODULE"""
    db, _ = parser.parseFromString(DATA)
    tdef = TypedefStructure.get(db.session, "EcuTask")
    assert tdef.name == "EcuTask"
    assert tdef.longIdentifier == "TYPEDEF for class EcuTask"
    assert tdef.size == 88
    assert tdef.link == "SYMBOL_TYPE_LINK"
    assert tdef.symbol == "EcuTask"
    components = tdef.components
    assert len(components) == 6
    comp = components[0]
    assert comp.name == "taskId"
    assert comp.offset == 0
    assert comp.link == "SYMBOL_TYPE_LINK"
    assert comp.symbol == "taskId"
    comp = components[1]
    assert comp.name == "counter"
    assert comp.offset == 2
    assert comp.link == "SYMBOL_TYPE_LINK"
    assert comp.symbol == "counter"
    comp = components[2]
    assert comp.name == "channel1"
    assert comp.offset == 16
    assert comp.link == "SYMBOL_TYPE_LINK"
    assert comp.symbol == "channel1"
    comp = components[3]
    assert comp.name == "byte"
    assert comp.offset == 0x38
    assert comp.link == "SYMBOL_TYPE_LINK"
    assert comp.symbol == "byte"
    comp = components[4]
    assert comp.name == "word"
    assert comp.offset == 0x3A
    assert comp.link == "SYMBOL_TYPE_LINK"
    assert comp.symbol == "word"
    comp = components[5]
    assert comp.name == "dword"
    assert comp.offset == 0x3C
    assert comp.link == "SYMBOL_TYPE_LINK"
    assert comp.symbol == "dword"
    assert len(tdef.instances) == 3
    instances = tdef.instances
    inst = instances[0]
    assert inst.name == "ecuTask1"
    assert inst.longIdentifier == "ecupp task number 1"
    assert inst.typeName == "EcuTask"
    assert inst.address == 0x00000000
    inst = instances[1]
    assert inst.name == "ecuTask2"
    assert inst.longIdentifier == "ecu task number 2"
    assert inst.typeName == "EcuTask"
    assert inst.address == 0x00000000
    inst = instances[2]
    assert inst.name == "activeEcuTask"
    assert inst.longIdentifier == "pointer to active ecu task"
    assert inst.typeName == "EcuTask"
    assert inst.address == 0x00000000


@pytest.mark.skip
def test_compu_method_invalid():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.TAB_VERB.DEFAULT_VALUE
          "Verbal conversion with default value"
          FOO_BAR "%12.0" ""
        /end COMPU_METHOD
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    CompuMethod(db.session, module.compu_method[0].name)


def test_compu_method_tab_verb():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.TAB_VERB.DEFAULT_VALUE
          "Verbal conversion with default value"
          TAB_VERB "%12.0" ""
          COMPU_TAB_REF CM.TAB_VERB.DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_VTAB CM.TAB_VERB.DEFAULT_VALUE.REF
          "List of text strings and relation to impl value"
          TAB_VERB 3
          1 "SawTooth"
          2 "Square"
          3 "Sinus"
          DEFAULT_VALUE "unknown signal type"
        /end COMPU_VTAB
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    assert compu.int_to_physical(1) == "SawTooth"
    assert compu.physical_to_int("Sinus") == 3
    assert compu.int_to_physical(10) == "unknown signal type"


def test_compu_method_tab_verb_no_default_value():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.TAB_VERB.DEFAULT_VALUE
          "Verbal conversion with default value"
          TAB_VERB "%12.0" ""
          COMPU_TAB_REF CM.TAB_VERB.DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_VTAB CM.TAB_VERB.DEFAULT_VALUE.REF
          "List of text strings and relation to impl value"
          TAB_VERB 3
          1 "SawTooth"
          2 "Square"
          3 "Sinus"
        /end COMPU_VTAB
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    assert compu.int_to_physical(1) == "SawTooth"
    assert compu.physical_to_int("Sinus") == 3
    assert compu.int_to_physical(10) is None


@pytest.mark.skip
def test_compu_method_tab_verb_no_vtab():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.TAB_VERB.DEFAULT_VALUE
          "Verbal conversion with default value"
          TAB_VERB "%12.0" ""
          COMPU_TAB_REF CM.TAB_VERB.DEFAULT_VALUE.REF
        /end COMPU_METHOD
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    with pytest.raises(exceptions.StructuralError):
        CompuMethod(db.session, module.compu_method[0])


def test_compu_method_tab_nointerp_default():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.TAB_NOINTP.DEFAULT_VALUE
          ""
          TAB_NOINTP "%8.4" "U/  min  "
          COMPU_TAB_REF CM.TAB_NOINTP.DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_TAB CM.TAB_NOINTP.DEFAULT_VALUE.REF
           ""
           TAB_NOINTP
           12
           -3 98
           -1 99
           0 100
           2 102
           4 104
           5 105
           6 106
           7 107
           8 108
           9 109
           10 110
           13 111
           DEFAULT_VALUE_NUMERIC 300.56 /* DEFAULT_VALUE_NUME RIC should be used here as the normal output is numeric */
        /end COMPU_TAB
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    assert compu.int_to_physical(-3) == 98
    assert compu.int_to_physical(8) == 108
    assert compu.physical_to_int(108) == 8
    assert compu.int_to_physical(1) == 300.56


def test_compu_method_tab_interp_default():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.TAB_INTP.DEFAULT_VALUE
          ""
          TAB_INTP "%8.4" "U/  min  "
          COMPU_TAB_REF CM.TAB_INTP.DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_TAB CM.TAB_INTP.DEFAULT_VALUE.REF
           ""
           TAB_INTP
           12
           -3 98
           -1 99
           0 100
           2 102
           4 104
           5 105
           6 106
           7 107
           8 108
           9 109
           10 110
           13 111
           DEFAULT_VALUE_NUMERIC 300.56 /* DEFAULT_VALUE_NUME RIC should be used here as the normal output is numeric */
        /end COMPU_TAB
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    xs = np.arange(-3, 14)
    ys = np.array(
        [
            98.0,
            98.5,
            99.0,
            100.0,
            101.0,
            102.0,
            103.0,
            104.0,
            105.0,
            106.0,
            107.0,
            108.0,
            109.0,
            110.0,
            110.33333333333333,
            110.66666666666667,
            111.0,
        ]
    )
    assert np.array_equal(compu.int_to_physical(xs), ys)
    assert compu.int_to_physical(-3) == 98
    assert compu.int_to_physical(8) == 108
    assert compu.int_to_physical(14) == 300.56
    assert compu.int_to_physical(-4) == 300.56


def test_compu_method_tab_interp_no_default():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.TAB_INTP.NO_DEFAULT_VALUE
          ""
          TAB_INTP "%8.4" "U/  min  "
          COMPU_TAB_REF CM.TAB_INTP.NO_DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_TAB CM.TAB_INTP.NO_DEFAULT_VALUE.REF
           ""
           TAB_INTP
           12
           -3 98
           -1 99
           0 100
           2 102
           4 104
           5 105
           6 106
           7 107
           8 108
           9 109
           10 110
           13 111
        /end COMPU_TAB
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    xs = np.arange(-3, 14)
    ys = np.array(
        [
            98.0,
            98.5,
            99.0,
            100.0,
            101.0,
            102.0,
            103.0,
            104.0,
            105.0,
            106.0,
            107.0,
            108.0,
            109.0,
            110.0,
            110.33333333333333,
            110.66666666666667,
            111.0,
        ]
    )
    assert np.array_equal(compu.int_to_physical(xs), ys)
    assert compu.int_to_physical(-3) == 98
    assert compu.int_to_physical(8) == 108
    assert compu.int_to_physical(14) is None
    assert compu.int_to_physical(-4) is None


def test_compu_method_tab_nointerp_no_default():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.TAB_NOINTP.NO_DEFAULT_VALUE
          ""
          TAB_NOINTP "%8.4" "U/  min  "
          COMPU_TAB_REF CM.TAB_NOINTP.NO_DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_TAB CM.TAB_NOINTP.NO_DEFAULT_VALUE.REF
           ""
           TAB_NOINTP
           12
           -3 98
           -1 99
           0 100
           2 102
           4 104
           5 105
           6 106
           7 107
           8 108
           9 109
           10 110
           13 111
        /end COMPU_TAB
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    assert compu.int_to_physical(-3) == 98
    assert compu.int_to_physical(8) == 108
    assert compu.physical_to_int(108) == 8
    assert compu.int_to_physical(1) is None


@pytest.mark.skip
def test_compu_method_tab_nointerp_both_defaults():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.TAB_NOINTP.DEFAULT_VALUE
          ""
          TAB_NOINTP "%8.4" "U/  min  "
          COMPU_TAB_REF CM.TAB_NOINTP.DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_TAB CM.TAB_NOINTP.DEFAULT_VALUE.REF
           ""
           TAB_NOINTP
           12
           -3 98
           -1 99
           0 100
           2 102
           4 104
           5 105
           6 106
           7 107
           8 108
           9 109
           10 110
           13 111
           DEFAULT_VALUE "value out of range"
           DEFAULT_VALUE_NUMERIC 300.56 /* DEFAULT_VALUE_NUME RIC should be used here as the normal output is numeric */
        /end COMPU_TAB
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    with pytest.raises(exceptions.StructuralError):
        CompuMethod(db.session, module.compu_method[0])


def test_compu_method_tab_verb_ranges():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.VTAB_RANGE.DEFAULT_VALUE
           "verbal range with default value"
           TAB_VERB
           "%4.2"
           ""
           COMPU_TAB_REF CM.VTAB_RANGE.DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_VTAB_RANGE CM.VTAB_RANGE.DEFAULT_VALUE.REF
           ""
           11
           0 1 "Zero_to_one"
           2 3 "two_to_three"
           4 7 "four_to_seven"
           14 17 "fourteen_to_seventeen"
           18 99 "eigteen_to_ninetynine"
           100 100 "hundred"
           101 101 "hundredone"
           102 102 "hundredtwo"
           103 103 "hundredthree"
           104 104 "hundredfour"
           105 105 "hundredfive"
           DEFAULT_VALUE "out of range value"
        /end COMPU_VTAB_RANGE
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    assert compu.int_to_physical(0) == "Zero_to_one"
    assert compu.int_to_physical(6) == "four_to_seven"
    assert compu.int_to_physical(45) == "eigteen_to_ninetynine"
    assert compu.int_to_physical(100) == "hundred"
    assert compu.int_to_physical(105) == "hundredfive"
    assert compu.int_to_physical(-1) == "out of range value"
    assert compu.int_to_physical(106) == "out of range value"
    assert compu.int_to_physical(10) == "out of range value"


def test_compu_method_tab_verb_ranges_no_default():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.VTAB_RANGE.NO_DEFAULT_VALUE
           "verbal range without default value"
           TAB_VERB
           "%4.2"
           ""
           COMPU_TAB_REF CM.VTAB_RANGE.NO_DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_VTAB_RANGE CM.VTAB_RANGE.NO_DEFAULT_VALUE.REF
           ""
           11
           0 1 "Zero_to_one"
           2 3 "two_to_three"
           4 7 "four_to_seven"
           14 17 "fourteen_to_seventeen"
           18 99 "eigteen_to_ninetynine"
           100 100 "hundred"
           101 101 "hundredone"
           102 102 "hundredtwo"
           103 103 "hundredthree"
           104 104 "hundredfour"
           105 105 "hundredfive"
        /end COMPU_VTAB_RANGE
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    assert compu.int_to_physical(0) == "Zero_to_one"
    assert compu.int_to_physical(6) == "four_to_seven"
    assert compu.int_to_physical(45) == "eigteen_to_ninetynine"
    assert compu.int_to_physical(100) == "hundred"
    assert compu.int_to_physical(105) == "hundredfive"
    assert compu.int_to_physical(-1) is None
    assert compu.int_to_physical(106) is None
    assert compu.int_to_physical(10) is None


def test_compu_method_tab_verb_ranges_inv():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.VTAB_RANGE.DEFAULT_VALUE
           "verbal range with default value"
           TAB_VERB
           "%4.2"
           ""
           COMPU_TAB_REF CM.VTAB_RANGE.DEFAULT_VALUE.REF
        /end COMPU_METHOD
        /begin COMPU_VTAB_RANGE CM.VTAB_RANGE.DEFAULT_VALUE.REF
           ""
           11
           0 1 "Zero_to_one"
           2 3 "two_to_three"
           4 7 "four_to_seven"
           14 17 "fourteen_to_seventeen"
           18 99 "eigteen_to_ninetynine"
           100 100 "hundred"
           101 101 "hundredone"
           102 102 "hundredtwo"
           103 103 "hundredthree"
           104 104 "hundredfour"
           105 105 "hundredfive"
           DEFAULT_VALUE "out of range value"
        /end COMPU_VTAB_RANGE
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    assert compu.physical_to_int("Zero_to_one") == 0
    assert compu.physical_to_int("four_to_seven") == 4
    assert compu.physical_to_int("eigteen_to_ninetynine") == 18
    assert compu.physical_to_int("hundred") == 100
    assert compu.physical_to_int("hundredfive") == 105
    assert compu.physical_to_int("out of range value") is None


def test_compu_method_identical():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.IDENTICAL
          "conversion that delivers always phys = int"
          IDENTICAL "%3.0" "hours"
        /end COMPU_METHOD
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    xs = np.arange(-10, 11)
    assert np.array_equal(compu.int_to_physical(xs), xs)
    assert np.array_equal(compu.physical_to_int(xs), xs)


@pytest.mark.skipif("RUN_MATH_TEST == False")
def test_compu_method_rat_func_identical():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.RAT_FUNC.IDENT
          "rational function with parameter set for int = f(phys) = phys"
          RAT_FUNC "%3.1" "m/s"
          COEFFS 0 1 0 0 0 1
        /end COMPU_METHOD
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    xs = np.arange(-10, 11)
    assert np.array_equal(compu.int_to_physical(xs), xs)
    assert np.array_equal(compu.physical_to_int(xs), xs)


@pytest.mark.skipif("RUN_MATH_TEST == False")
def test_compu_method_rat_func_linear():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.RAT_FUNC.DIV_81_9175
          "rational function with parameter set for impl = f(phys) = phys * 81.9175"
          RAT_FUNC "%8.4" "grad C"
          COEFFS 0 81.9175 0 0 0 1
        /end COMPU_METHOD
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    xs = np.arange(-10, 11)
    ys = np.array(
        [
            -819.1750000000001,
            -737.2575,
            -655.34,
            -573.4225,
            -491.505,
            -409.58750000000003,
            -327.67,
            -245.7525,
            -163.835,
            -81.9175,
            0.0,
            81.9175,
            163.835,
            245.7525,
            327.67,
            409.58750000000003,
            491.505,
            573.4225,
            655.34,
            737.2575,
            819.1750000000001,
        ]
    )
    assert np.array_equal(compu.int_to_physical(ys), xs)
    assert np.array_equal(compu.physical_to_int(xs), ys)


@pytest.mark.skip
@pytest.mark.skipif("RUN_MATH_TEST == False")
def test_compu_method_rat_func_no_coeffs():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.RAT_FUNC.DIV_81_9175
          "rational function with parameter set for impl = f(phys) = phys * 81.9175"
          RAT_FUNC "%8.4" "grad C"
        /end COMPU_METHOD
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    with pytest.raises(exceptions.StructuralError):
        CompuMethod(db.session, module.compu_method[0])


@pytest.mark.skipif("RUN_MATH_TEST == False")
def test_compu_method_linear():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.LINEAR.MUL_2
        "Linear function with parameter set for phys = f(int) = 2*int + 0"
         LINEAR "%3.1" "m/s"
         COEFFS_LINEAR 2 0
        /end COMPU_METHOD
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    xs = np.arange(-10, 11)
    assert np.array_equal(compu.int_to_physical(xs), xs * 2.0)
    assert np.array_equal(compu.physical_to_int(xs * 2.0), xs)


@pytest.mark.skip
@pytest.mark.skipif("RUN_MATH_TEST == False")
def test_compu_method_linear_no_coeffs():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.LINEAR.MUL_2
        "Linear function with parameter set for phys = f(int) = 2*int + 0"
         LINEAR "%3.1" "m/s"
        /end COMPU_METHOD
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    with pytest.raises(exceptions.StructuralError):
        CompuMethod(db.session, module)


@pytest.mark.skipif("RUN_MATH_TEST == False")
def test_compu_method_formula_with_inv():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.FORM.X_PLUS_4
          ""
          FORM
          "%6.1"
          "rpm"
          /begin FORMULA
            "X1+4"
            FORMULA_INV "X1-4"
          /end FORMULA
        /end COMPU_METHOD
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    assert compu.int_to_physical(6) == 10
    assert compu.physical_to_int(4) == 0


@pytest.mark.skipif("RUN_MATH_TEST == False")
def test_compu_method_formula_without_inv():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin COMPU_METHOD CM.FORM.X_PLUS_4
          ""
          FORM
          "%6.1"
          "rpm"
          /begin FORMULA
            "X1+4"
          /end FORMULA
        /end COMPU_METHOD
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    assert compu.int_to_physical(6) == 10


@pytest.mark.skipif("RUN_MATH_TEST == False")
def test_compu_method_formula_with_sysc():
    parser = ParserWrapper("a2l", "module", A2LListener)
    DATA = """
    /begin MODULE testModule ""
        /begin MOD_PAR ""
             SYSTEM_CONSTANT "System_Constant_1" "42"
             SYSTEM_CONSTANT "System_Constant_2" "Textual constant"
        /end MOD_PAR

        /begin COMPU_METHOD CM.FORM.X_PLUS_SYSC
          ""
          FORM
          "%6.1"
          "rpm"
          /begin FORMULA
            "X1 + sysc(System_Constant_1)"
          /end FORMULA
        /end COMPU_METHOD
    /end MODULE
    """
    db, _ = parser.parseFromString(DATA)
    module = db.session.query(model.Module).first()
    compu = CompuMethod(db.session, module.compu_method[0].name)
    assert compu.int_to_physical(23) == 65



#!/usr/bin/env python
# -*- coding: latin-1 -*-

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

"""Codes for scaling units (CSE)
"""

from collections import namedtuple
from enum import IntEnum

CSE_Type = namedtuple("CSE_Type", "code unit referred_to comment")


class Referer(IntEnum):
    """"""

    NONE = 0
    TIME = 1
    ANGLE = 2
    COMBUSTION = 3
    EVENT = 4


CSE = {
    0: CSE_Type(0, "1 µsec", Referer.TIME, ""),
    1: CSE_Type(1, "10 µsec", Referer.TIME, ""),
    2: CSE_Type(2, "100 µsec", Referer.TIME, ""),
    3: CSE_Type(3, "1 msec", Referer.TIME, ""),
    4: CSE_Type(4, "10 msec", Referer.TIME, ""),
    5: CSE_Type(5, "100 msec", Referer.TIME, ""),
    6: CSE_Type(6, "1 sec", Referer.TIME, ""),
    7: CSE_Type(7, "10 sec", Referer.TIME, ""),
    8: CSE_Type(8, "1 min", Referer.TIME, ""),
    9: CSE_Type(9, "1 hour", Referer.TIME, ""),
    10: CSE_Type(10, "1 day", Referer.TIME, ""),
    100: CSE_Type(100, "Angular degrees", Referer.ANGLE, ""),
    101: CSE_Type(101, "Revolutions 360 degrees", Referer.ANGLE, ""),
    102: CSE_Type(102, "Cycle 720 degrees", Referer.ANGLE, "e.g. in case of IC engines"),
    103: CSE_Type(103, "Cylinder segment", Referer.COMBUSTION, "e.g. in case of IC engines"),
    998: CSE_Type(998, "When frame available", Referer.EVENT, "Source defined in keyword Frame"),
    999: CSE_Type(
        999,
        "Always if there is new value",
        Referer.NONE,
        """Calculation of a new
                     upper range limit after receiving a new partial value, e.g. when
                     calculating a complex trigger condition""",
    ),
    1000: CSE_Type(1000, "Non deterministic", Referer.NONE, "Without fixed scaling"),
}



#!/usr/bin/env python

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2025 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.10.2"

import sys
import typing
from io import TextIOWrapper


try:
    from rich.traceback import install

    RICH_AVAILABLE = True
except ImportError:
    RICH_AVAILABLE = False

import pya2l.model as model
from pya2l.a2lparser import A2LParser, path_components
from pya2l.logger import Logger
from pya2l.templates import doTemplateFromText


if RICH_AVAILABLE:
    install(show_locals=True, max_frames=3)  # Install custom exception handler.
pyver = sys.version_info


class InvalidA2LDatabase(Exception):
    """"""

    pass


a2l_logger = Logger("A2LDB", "INFO")

if pyver.major == 3 and pyver.minor <= 9:
    import pkg_resources

    A2L_TEMPLATE = str(pkg_resources.resource_string("pya2l.cgen.templates", "a2l.tmpl"), encoding="utf8")
else:
    import importlib.resources

    with (importlib.resources.files("pya2l.cgen.templates") / "a2l.tmpl").open(encoding="utf8") as data:
        A2L_TEMPLATE = data.read()


def import_a2l(
    file_name: str,
    debug: bool = False,
    in_memory: bool = False,
    remove_existing: bool = False,
    local: bool = False,
    encoding: str = "latin-1",
    loglevel: str = "INFO",
    progress_bar: bool = True,
):
    """Import `.a2l` file to `.a2ldb` database.


    Parameters
    ----------
    file_name: str
        Name of the A2L to be imported. If you don't specify an extension ``.a2l`` is added.

    debug: bool
        Additional debugging output.

    in_memory: bool
        Create non-persistent in-memory database.

    remove_existing: bool
        ** DANGER ZONE **: Remove existing database.

    local: bool
        If `True` create A2LDB in current working directory else in A2L source directory.

    encoding: str
        File encoding like "latin-1" or "utf-8" or None to auto-detect.

    loglevel: str
        "INFO" | "WARN" | "DEBUG" | "ERROR" | "CRITICAL"

    progress_bar: bool
        Disable progress bar.

    Returns
    -------
    SQLAlchemy session object.

    Raises
    ------
    OSError
        If database already exists.

    Note
    ----
    ``AML`` and ``IF_DATA`` sections are currently not processed.
    """

    a2l_parser = A2LParser()
    db = a2l_parser.parse(
        file_name=file_name,
        local=local,
        in_memory=in_memory,
        encoding=encoding,
        remove_existing=remove_existing,
        loglevel=loglevel,
        progress_bar=progress_bar,
    )
    session = db.session
    session.commit()
    session.setup_ifdata_parser(loglevel=loglevel)
    return session


def open_existing(file_name: str, loglevel: str = "INFO"):
    """Open an existing `.a2ldb` database.

    Parameters
    ----------
    file_name: str
        Name of your database file, resulting from :meth:`import_a2l`.
        Extension `.a2ldb` not needed.

    Returns
    -------
    SQLAlchemy session object.

    Raises
    ------
    OSError
        If database already exists.
    """
    _, db_fn = path_components(in_memory=False, file_name=file_name)

    if not db_fn.exists():
        raise OSError(f"file {db_fn!r} does not exists.")
    else:
        db = model.A2LDatabase(str(db_fn))
        session = db.session
        res = session.query(model.MetaData).first()
        if res:
            session.setup_ifdata_parser(loglevel=loglevel)
            return session
        else:
            raise InvalidA2LDatabase("Database seems to be corrupted. No meta-data found.")
        return session


def open_create(file_name: str, local: bool = False, encoding: str = "latin-1", loglevel: str = "INFO"):
    """Open or create an A2LDB."""

    a2l_fn, db_fn = path_components(in_memory=False, file_name=file_name)
    if not db_fn.exists():
        return import_a2l(a2l_fn, local=local, encoding=encoding, loglevel=loglevel)
    else:
        return open_existing(db_fn, loglevel)


def export_a2l(db_name: str, output: typing.Union[TextIOWrapper, str, typing.Any] = sys.stdout, encoding="latin1"):  # noqa: UP007
    """
    Parameters
    ----------
    file_name: str
        Name of the A2L exported.

    encoding: str
        File encoding like "latin-1" or "utf-8".
    """
    session = open_existing(db_name)
    namespace = dict(session=session, model=model)
    data = doTemplateFromText(A2L_TEMPLATE, namespace, formatExceptions=False, encoding=encoding)
    if isinstance(output, TextIOWrapper):
        output.write(data)
    else:
        with open(file=output, mode="w", encoding=encoding) as outf:
            outf.write(data)


class DB:
    """"""

    def __init__(self, loglevel: str = "") -> None:
        if loglevel:
            a2l_logger.setLevel(loglevel)

    @staticmethod
    def import_a2l(
        file_name: str,
        debug: bool = False,
        in_memory: bool = False,
        remove_existing: bool = False,
        local: bool = False,
        encoding: str = "latin-1",
        loglevel: str = "INFO",
        progress_bar: bool = True,
    ):
        return import_a2l(file_name, debug, in_memory, remove_existing, local, encoding, loglevel, progress_bar)

    @staticmethod
    def open_existing(file_name: str, loglevel: str = "INFO"):
        return open_existing(file_name, loglevel)

    @staticmethod
    def open_create(file_name: str, local: bool = False, encoding: str = "latin-1", loglevel: str = "INFO"):
        return open_create(file_name, local, encoding, loglevel)

    @staticmethod
    def export_a2l(
        db_name: str, output: typing.Union[TextIOWrapper, str, typing.Any] = sys.stdout, encoding: str = "latin1"  # noqa: UP007
    ):
        export_a2l(db_name, output, encoding)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2020 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


from pya2l.logger import Logger


class MixInBase:

    logger = Logger(__name__)


class AxisDescrMixIn(MixInBase):
    """"""

    def check(self):
        if self.attribute == "CURVE_AXIS":
            if self.conversion != "NO_COMPU_METHOD":
                self.logger.error("CURVE_AXIS have no input conversion, use 'NO_COMPU_METHOD' for argument 'conversion'.")
                return False
        #            meas =
        return True


class CompareByPositionMixIn(MixInBase):
    """
    Enable sortability in user code.

    Implements basic comparison.
    """

    def __eq__(self, other):
        return self.position == other.position

    def __lt__(self, other):
        return self.position < other.position


MIXIN_MAP = {
    "AXIS_DESCR": "AxisDescrMixIn",
}



from pya2l import DB
from pya2l import model

db = DB()
session = db.open_create("ASAP2_Demo_V161")
measurements = session.query(model.Measurement).order_by(model.Measurement.name).all()
for m in measurements:
    print("{:48} {:12} 0x{:08x}".format(m.name, m.datatype, m.ecu_address.address))
# print(dir(m))
ifd = model.IfData()
print(ifd)
print(dir(ifd))
